import torch
from collections import Counter
from enum import Enum
from irisml.tasks.create_azure_openai_chat_model import OpenAITextChatModel
import logging
from torchmetrics import Metric

from visionmetrics.common.utils import precision_recall_f1_scalar

logger = logging.getLogger(__name__)


# Prompt template placeholder constants
PREDICTION_PLACEHOLDER = "<|prediction|>"
TARGET_PLACEHOLDER = "<|target|>"

# Response parsing constants
MULTIPLE_RESPONSES_DELIMITER = "<|delimiter|>"

# Evaluator separator constants for joining multiple ground truth values
OR_SEPARATOR = " <|OR|> "
AND_SEPARATOR = " <|AND|> "


# Enum for per-value result status type
class ResultStatusType(Enum):
    TruePositive = 1
    TrueNegative = 2
    FalseNegative = 3
    FalsePositiveGtNotNull = 4
    FalsePositiveGtNull = 5


class AzureOpenAITextModelCategoricalEvaluatorBase(Metric):
    """
    Evaluator that uses single-turn text-only calls to an Azure OpenAI model for evaluation and returns
    standard categorical metrics (precision, recall, F1, accuracy) based on the scores returned.

    Args:
        prompt_template: string template for the prompt instructing how to score the prediction. Must have two placeholders: <|prediction|> and <|target|>.
        positive_threshold: determines the threshold (inclusive) at which the score assigned by the model to the prediction means the prediction is correct. Range: [0.0, 1.0].
        negative_value: indicates the value (e.g., None, 'null', 0, '', []) for which predicted and target values are considered null. Note that the evaluator strictly requires
                        an exact match for either a prediction or target to be considered negative. TODO: Implement fuzzy negative matching. Scores are always computed so they
                        can be used for average_score computation.

        Other parameters follow the standards of irisml.tasks.create_azure_openai_chat_model.OpenAITextChatModel;
        see https://github.com/microsoft/irisml-tasks-azure-openai/blob/main/irisml/tasks/create_azure_openai_chat_model.py.
    """
    def __init__(self, endpoint: str, deployment_name: str, system_message: str, prompt_template: str,
                 temperature=0.0, max_tokens=50, requests_interval=0, num_responses=1, positive_threshold=0.5, negative_value='', api_version="2024-09-01-preview"):
        super().__init__()
        if PREDICTION_PLACEHOLDER not in prompt_template or TARGET_PLACEHOLDER not in prompt_template:
            raise ValueError("Both the predicted placeholder {PREDICTION_PLACEHOLDER} and target placeholder {TARGET_PLACEHOLDER} must be present in prompt_template.")
        if positive_threshold < 0.0 or positive_threshold > 1.0:
            raise ValueError("Parameter positive_threshold should be between [0.0, 1.0], inclusive.")
        self.system_message = system_message
        self.prompt_template = prompt_template
        self.positive_threshold = positive_threshold
        self.negative_value = negative_value

        if num_responses != 1:
            raise NotImplementedError("This metric currently only supports single-response scoring.")
        self.num_responses = num_responses

        self.model = OpenAITextChatModel(endpoint=endpoint, deployment_name=deployment_name, api_version=api_version, api_key=None, temperature=temperature, max_tokens=max_tokens,
                                         requests_interval=requests_interval, num_responses=num_responses, delimiter=MULTIPLE_RESPONSES_DELIMITER, system_message=system_message)

        self.add_state("predictions", default=[], dist_reduce_fx="cat")
        self.add_state("targets", default=[], dist_reduce_fx="cat")
        self.add_state("raw_scores", default=[], dist_reduce_fx="cat")
        self.add_state("scores", default=[], dist_reduce_fx="cat")
        self.add_state("score_parse_failures", default=torch.tensor(0), dist_reduce_fx="sum")
        self.add_state("result_status_types", default=[], dist_reduce_fx="cat")

        logger.info(f"Initialized AzureOpenAITextModelCategoricalEvaluatorBase with positive_threshold={positive_threshold}, negative_value={negative_value}, temperature={temperature}, "
                    f"max_tokens={max_tokens}, system_message=\"{system_message}\", prompt_template=\"{prompt_template}\"")

    def _get_numeric_score(self, raw_score: str):
        """
        Parses the numeric score value from the string `raw_score`; if not parseable, returns 0.
        """
        score = 0.
        try:
            score = float(raw_score)
        except ValueError:
            self.score_parse_failures += 1
            logger.debug(f"Failed to parse raw_score \"{raw_score}\" to numeric value; returning 0.")
        return score

    def update(self, predictions, targets, multi_target_separator=AND_SEPARATOR):
        """
        Evaluate list of predicted results using Azure OpenAI text model.
        Args:
            predictions: list of string predictions [text1, text2, ...], shape: (N, ), type: string
            targets: list of one or more string ground truth values: [[gt1, gt2, ...], [gt1, gt2, ...], ...], type: string
                Note: with multiple targets, a good match with any of them is considered correct.
            multiple_target_join_method: string specifying the joining method to join multiple reference ground truth values in the prompt, either 'any' or 'all'
        """
        if len(predictions) != len(targets):
            raise ValueError("Please update with an equal number of predictions and targets.")

        self.predictions += predictions
        self.targets += targets

        for prediction, target in zip(predictions, targets):
            prompt = self.prompt_template.replace(PREDICTION_PLACEHOLDER, prediction)
            final_target = multi_target_separator.join(target)
            prompt = prompt.replace(TARGET_PLACEHOLDER, final_target)

            if prediction == target:
                result = "1.0"
            else:
                result = self.model(([prompt], [[]]))[0]
            if self.num_responses == 1:
                self.raw_scores.append(result)
                numeric_score = self._get_numeric_score(result)
                self.scores.append(numeric_score)
            else:
                # TODO: Implement multiple-response score aggregation
                pass

            # Assign ResultStatusTypes
            target_is_negative = not target or all([t == self.negative_value for t in target])
            prediction_match = numeric_score >= self.positive_threshold
            if prediction == self.negative_value:
                # If any target value is negative, it is a true negative
                if target_is_negative:
                    self.result_status_types.append(ResultStatusType.TrueNegative)
                else:
                    self.result_status_types.append(ResultStatusType.FalseNegative)
            else:
                if target_is_negative:
                    self.result_status_types.append(ResultStatusType.FalsePositiveGtNull)
                else:
                    self.result_status_types.append(ResultStatusType.TruePositive if prediction_match else ResultStatusType.FalsePositiveGtNotNull)

    def compute(self):
        result_counts = Counter(self.result_status_types)
        tp, tn, fn, fp_gt_null, fp_gt_not_null = result_counts[ResultStatusType.TruePositive], \
            result_counts[ResultStatusType.TrueNegative], \
            result_counts[ResultStatusType.FalseNegative], \
            result_counts[ResultStatusType.FalsePositiveGtNull], \
            result_counts[ResultStatusType.FalsePositiveGtNotNull]
        precision_recall_f1 = precision_recall_f1_scalar(tp=tp, fp=fp_gt_null + fp_gt_not_null, fn=fn + fp_gt_not_null)
        try:
            accuracy = (tp + tn) / (tp + tn + fn + fp_gt_null + fp_gt_not_null)
        except ZeroDivisionError:
            accuracy = 0.
        try:
            average_score = sum(self.scores) / len(self.scores)
        except ZeroDivisionError:
            average_score = 0.
        return {
            # Summary statistics
            "Precision": precision_recall_f1["Precision"],
            "Recall": precision_recall_f1["Recall"],
            "F1": precision_recall_f1["F1"],
            "Accuracy": accuracy,
            "AverageScore": average_score,
            # Raw statistic counts
            "TP": tp,
            "TN": tn,
            "FN": fn,
            "FP_GTNull": fp_gt_null,
            "FP_GTNotNull": fp_gt_not_null,
            "ScoreParseFailures": self.score_parse_failures.item()
        }



# Import metrics directly from torchmetrics


# Import custom metrics from visionmetrics
from visionmetrics.key_value_pair.key_value_pair_eval import KeyValuePairExtractionScore


__all__ = ['KeyValuePairExtractionScore']



from enum import Enum
import logging

from visionmetrics.key_value_pair.key_value_pair_eval_base import KeyValuePairEvaluatorBase, SupportedKeyWiseMetric

logger = logging.getLogger(__name__)


class JSONSchemaKeyType(str, Enum):
    String = "string"
    Number = "number"
    Integer = "integer"
    Boolean = "boolean"
    Array = "array"
    Object = "object"


# Constants for schemas
SIMPLE_KEY_TYPES = [JSONSchemaKeyType.String, JSONSchemaKeyType.Number, JSONSchemaKeyType.Integer, JSONSchemaKeyType.Boolean]
OUT_OF_DISTRIBUTION_ENUM_KEY = "<|other|>"

# Constants for predictions and targets
VALUE_SUBKEY = "value"
GROUNDINGS_SUBKEY = "groundings"


class KeyValuePairExtractionScore(KeyValuePairEvaluatorBase):
    """
    Key-value pair extraction refers to the task of arbitrary schema-based structured field extraction. Each schema is an adapted version of a
    JSON Schema (https://json-schema.org/understanding-json-schema/reference)-formatted dictionary that contains keys, each of which specifies
    the standard JSON Schema type of the key and a string description, whether to perform grounding on the key, classes for closed-vocabulary
    values, and additional information describing list items and object properties (sub-keys).

    Exceptions to the standard JSON Schema format are:
    - Usage of an "includeGrounding" boolean for each field, indicating whether the field's annotations are grounded (whether each has a list of bounding boxes) or not.

    Based on the properties defined in the JSON Schema, this class infers the best evaluation metric for each key's data type, and defaults to text-based evaluation
    for cases that have no clear inferrable metric. It then constructs the key_metric_map specifying the mapping between each key of such a schema and the corresponding
    visionmetrics metric, metric arguments, preprocessors for predictions and targets, and a list trace describing the path to the key through the prediction and target objects.
    Evaluation and metric computation are done through the parent class, KeyValuePairEvaluatorBase. Supported key-wise metrics are enumerated in KeyValuePairEvaluatorBase.SupportedKeyWiseMetric.

    Args:
        key_value_pair_schema: dictionary in JSON Schema format indicating each key and expected value type for each extracted field.
        Examples (corresponding to the examples in key_value_pair_eval_base.py):
        defect_detection_schema = {
            "defects": {
                "type": "array",
                "items": {
                    "type": "string",
                    "enum": ["scratch", "dent", "discoloration", "crack"],
                    "includeGrounding": True
                }
            },
            "rationale": {
                "type": "string"
            }
        }
        "brand_sentiment": {
            "type": "object",
            "properties": {
                "has_non_contoso_brands": {
                    "type": "boolean"
                },
                "contoso_specific": {
                    "type": "object",
                    "properties": {
                        "sentiment": {
                            "type": "string",
                            "enum": ["very positive", "somewhat positive", "neutral", "somewhat negative", "negative"]
                        },
                        "logos": {
                            "type": "array",
                            "items": {
                                "type": "string",
                                "enum": ["text", "grayscale", "rgb"],
                                "includeGrounding": True
                            }
                        }
                    }
                }
            }
        }
        Note: In the above brand_sentiment_schema example, since the nested objects are all objects (not arrays), the object is recursively traversed to assign metrics to the
        innermost object fields ("sentiment" and "logos").

        endpoint: string of the Azure OpenAI endpoint to be used as the default text evaluator.
        deployment_name: string of the Azure OpenAI deployment name to be used for the default text evaluator.
        api_version: string of the API version to use with the Azure OpenAI endpoint and deployment.

        The latter two arguments follow the standards of irisml.tasks.create_azure_openai_chat_model.OpenAITextChatModel;
        see https://github.com/microsoft/irisml-tasks-azure-openai/blob/main/irisml/tasks/create_azure_openai_chat_model.py.
    """
    def __init__(self, key_value_pair_schema: dict, endpoint: str, deployment_name: str, api_version: str = "2024-09-01-preview"):
        if not isinstance(key_value_pair_schema, dict):
            raise ValueError("key_value_pair_schema must be a dictionary in JSON Schema format specifying the schema for the dataset.")
        self.key_value_pair_schema = key_value_pair_schema
        self.endpoint = endpoint
        self.deployment_name = deployment_name
        self.api_version = api_version

        # Parse the schema and map each key to a metric
        self.key_metric_map = {}
        for key in key_value_pair_schema:
            key_schema = key_value_pair_schema[key]
            self._populate_key_metric_map(key=key, key_schema=key_schema, key_trace=[key])

        super().__init__(key_metric_map=self.key_metric_map)
        logger.info(f"Initialized KeyValuePairExtractionScore with schema {key_value_pair_schema}, endpoint {endpoint}, and deployment_name {deployment_name}.")

    def _get_enum_class_map(self, classes: list):
        """
        Constructs and returns an enum class map from string class names to integer class indices,
        used in preprocessing enum-type values that are evaluated using classification and detection metrics.
        Args:
            classes: list of class names. Class names can be strings, integers, or floats.
        """
        class_map = {str(class_name): i for i, class_name in enumerate(classes)}
        # Reserve one class to catch cases where predictions are not in the set of expected classes
        class_map[OUT_OF_DISTRIBUTION_ENUM_KEY] = len(class_map)
        return class_map

    def _assign_key_metric_map_values(self, key: str, metric_name: str, metric_args: dict, class_map: dict = None):
        """
        Assigns values for the metric_name, metric_args, and class_map for a given key in the object-level key_metric_map dictionary.
        Args:
            key: string of the key name, which should be identical to the key that will be evaluated in the prediction and target dictionaries.
            metric_name: string of the metric name as defined in visionmetrics (e.g., 'classification.MulticlassAccuracy'), which
                         should be among those specified in key_value_pair_eval_base.SupportedKeyWiseMetric.
            metric_args: dictionary of args to pass in as keyword arguments to the initialization function of the metric.
            class_map (optional): dictionary mapping string class names to integer class indices for this key.
        """
        if key not in self.key_metric_map:
            self.key_metric_map[key] = {}
        self.key_metric_map[key]["metric_name"] = metric_name
        self.key_metric_map[key]["metric_args"] = metric_args
        if class_map:
            self.key_metric_map[key]["class_map"] = class_map

    def _populate_key_metric_map(self, key: str, key_schema: dict, key_trace: list):
        """
        Recursive function that populates the object-level key_metric_map dictionary. For a given key and key_schema in JSON Schema format,
        the function populates the metric_name, metric_args, and preprocessor values in the dictionary.
        Note: The recursive behavior currently only occurs for JSONSchemaKeyType.Object key types, not arrays.
        Args:
            key: string of the key name, which should be identical to the key that will be evaluated in the prediction and target dictionaries.
            key_schema: dictionary in JSON schema format specifying the expected schema for the prediction and target dictionaries to be used in evaluation.
            key_trace: list of strings of key names that traces the path to the current key in the key-value pair prediction/target object (not in the schema).
        """
        if key_schema["type"] in [JSONSchemaKeyType.String, JSONSchemaKeyType.Number, JSONSchemaKeyType.Integer]:
            if "enum" in key_schema:
                class_map = self._get_enum_class_map(key_schema["enum"])
                if "includeGrounding" in key_schema and key_schema["includeGrounding"] is True:
                    self._assign_key_metric_map_values(key=key,
                                                       metric_name=SupportedKeyWiseMetric.Detection_MicroPrecisionRecallF1,
                                                       metric_args={"iou_threshold": 0.5, "box_format": "xyxy", "coords": "absolute"},
                                                       class_map=class_map)
                else:
                    self._assign_key_metric_map_values(key=key,
                                                       metric_name=SupportedKeyWiseMetric.Classification_MulticlassF1,
                                                       metric_args={"num_classes": len(class_map), "average": "micro"},
                                                       class_map=class_map)
            elif key_schema["type"] in [JSONSchemaKeyType.Number, JSONSchemaKeyType.Integer]:
                self._assign_key_metric_map_values(key=key,
                                                   metric_name=SupportedKeyWiseMetric.Regression_MeanAbsoluteErrorF1Score,
                                                   metric_args={"error_threshold": 0.0})
        elif key_schema["type"] == JSONSchemaKeyType.Boolean:
            class_map = self._get_enum_class_map([True, False])
            if "includeGrounding" in key_schema and key_schema["includeGrounding"] is True:
                self._assign_key_metric_map_values(key=key,
                                                   metric_name=SupportedKeyWiseMetric.Detection_MicroPrecisionRecallF1,
                                                   metric_args={"iou_threshold": 0.5, "box_format": "xyxy", "coords": "absolute"},
                                                   class_map=class_map)
            else:
                self._assign_key_metric_map_values(key=key,
                                                   metric_name=SupportedKeyWiseMetric.Classification_MulticlassF1,
                                                   metric_args={"num_classes": len(class_map), "average": "micro"},
                                                   class_map=class_map)
        elif key_schema["type"] == JSONSchemaKeyType.Array:
            # For more complex arrays, we default to the caption evaluator
            if key_schema["items"]["type"] in SIMPLE_KEY_TYPES:
                if "enum" in key_schema["items"]:
                    class_map = self._get_enum_class_map(key_schema["items"]["enum"])
                    if "includeGrounding" in key_schema["items"] and key_schema["items"]["includeGrounding"] is True:
                        self._assign_key_metric_map_values(key=key,
                                                           metric_name=SupportedKeyWiseMetric.Detection_MicroPrecisionRecallF1,
                                                           metric_args={"iou_threshold": 0.5, "box_format": "xyxy", "coords": "absolute"},
                                                           class_map=class_map)
                    else:
                        self._assign_key_metric_map_values(key=key,
                                                           metric_name=SupportedKeyWiseMetric.Classification_MultilabelF1WithDuplicates,
                                                           metric_args={},
                                                           class_map=class_map)
        elif key_schema["type"] == JSONSchemaKeyType.Object:
            for subkey in key_schema["properties"]:
                subkey_name = f"{key}_{subkey}"
                self._populate_key_metric_map(key=subkey_name, key_schema=key_schema["properties"][subkey], key_trace=key_trace + [VALUE_SUBKEY, subkey])

        if key not in self.key_metric_map and key_schema["type"] != JSONSchemaKeyType.Object:
            # Use text as the default metric for all keys; 'object' key type should not have its own key
            self._assign_key_metric_map_values(key=key,
                                               metric_name=SupportedKeyWiseMetric.Caption_AzureOpenAITextModelCategoricalScore,
                                               metric_args={"endpoint": self.endpoint, "deployment_name": self.deployment_name, "api_version": self.api_version})
            logger.debug(f"Using default metric '{SupportedKeyWiseMetric.Caption_AzureOpenAITextModelCategoricalScore.value}' for key '{key}'.")

        if key in self.key_metric_map:
            self._populate_key_preprocessor(key=key, type=key_schema["type"])
            self.key_metric_map[key]["key_trace"] = key_trace

    def _detection_preprocess_single_prediction(self, key: str, value: dict):
        """
        Preprocessing function for a single detection prediction of the form {"value": <class_name>, "groundings": [[[optional: <score>], <x1>, <y1>, <x2>, <y2>], [...], [...]]}.
        Args:
            key: string of the key name.
            value: dictionary with a "value" key of type string, integer, or float, and a "groundings" list of lists of bounding boxes, optionally with scores.
        """
        class_map = self.key_metric_map[key]["class_map"]
        class_index = class_map.get(str(value[VALUE_SUBKEY]), class_map.get(OUT_OF_DISTRIBUTION_ENUM_KEY))
        groundings = value[GROUNDINGS_SUBKEY]
        if not isinstance(groundings, list):
            return_pred = [[class_index] + [0.0] + [0., 0., 0., 0.]]
        else:
            return_pred = []
            for grounding in groundings:
                if not isinstance(grounding, list) or (isinstance(grounding, list) and len(grounding) < 4):
                    return_pred.append([class_index] + [0.0] + [0., 0., 0., 0.])
                else:
                    return_pred.append([class_index] + grounding if len(grounding) == 5 else [class_index] + [1.0] + grounding)
        return return_pred

    def _detection_preprocess_single_target(self, key: str, value: dict):
        """
        Preprocessing function for a single detection target of the form {"value": <class_name>, "groundings": [[<x1>, <y1>, <x2>, <y2>], [...], [...]]}.
        Args:
            key: string of the key name.
            value: dictionary with a "value" key of type string, integer, or float, and a "groundings" list of lists of bounding boxes.
        """
        class_map = self.key_metric_map[key]["class_map"]
        class_index = class_map.get(str(value[VALUE_SUBKEY]), class_map.get(OUT_OF_DISTRIBUTION_ENUM_KEY))
        groundings = value[GROUNDINGS_SUBKEY]
        if not isinstance(groundings, list):
            return_gt = [[0, 0., 0., 0., 0.]]
        else:
            return_gt = []
            for grounding in groundings:
                if not isinstance(grounding, list) or (isinstance(grounding, list) and len(grounding) < 4):
                    return_gt.append([class_index] + [0., 0., 0., 0.])
                else:
                    return_gt.append([class_index] + grounding)
        return return_gt

    def _populate_key_preprocessor(self, key: str, type: JSONSchemaKeyType):
        """
        Given a key and the specified JSON Schema type of that key, constructs the preprocessor function corresponding to the key
        and populates the object-level key_metric_map dictionary with the preprocessor function.
        Args:
            key: string of the key name, which should be identical to the key that will be evaluated in the prediction and target dictionaries.
            type: string specifying the JSON Schema type of the key, which should be among those in JSONSchemaKeyType.
        """
        metric_name = self.key_metric_map[key]["metric_name"]
        if metric_name == SupportedKeyWiseMetric.Caption_AzureOpenAITextModelCategoricalScore:
            # Expects list of strings for predictions, list of list of strings for targets
            self.key_metric_map[key]["prediction_preprocessor"] = lambda value: str(value[VALUE_SUBKEY])
            self.key_metric_map[key]["target_preprocessor"] = lambda value: [str(value[VALUE_SUBKEY])]
        elif metric_name == SupportedKeyWiseMetric.Classification_MulticlassAccuracy or metric_name == SupportedKeyWiseMetric.Classification_MulticlassF1:
            # Expects torch int or float tensor of shape (N, ...) or (N, C, ...) for predictions, torch tensor of shape (N, ...) for targets
            class_map = self.key_metric_map[key]["class_map"]
            self.key_metric_map[key]["prediction_preprocessor"] = self.key_metric_map[key]["target_preprocessor"] = \
                lambda value: class_map.get(str(value[VALUE_SUBKEY]), class_map.get(OUT_OF_DISTRIBUTION_ENUM_KEY))
        elif metric_name == SupportedKeyWiseMetric.Classification_MultilabelAccuracy or metric_name == SupportedKeyWiseMetric.Classification_MultilabelF1:
            # Expects torch int or float tensor of shape (N, C, ...)
            class_map = self.key_metric_map[key]["class_map"]

            def multilabel_preprocess(value: list):
                class_indices = [class_map.get(v[VALUE_SUBKEY], class_map.get(OUT_OF_DISTRIBUTION_ENUM_KEY)) for v in value[VALUE_SUBKEY]]
                one_hot_list = [1 if class_index in class_indices else 0 for class_index in range(0, len(class_map))]
                return one_hot_list
            self.key_metric_map[key]["prediction_preprocessor"] = self.key_metric_map[key]["target_preprocessor"] = multilabel_preprocess
        elif metric_name == SupportedKeyWiseMetric.Classification_MultilabelF1WithDuplicates:
            class_map = self.key_metric_map[key]["class_map"]
            self.key_metric_map[key]["prediction_preprocessor"] = self.key_metric_map[key]["target_preprocessor"] = \
                lambda value: [class_map.get(v[VALUE_SUBKEY], class_map.get(OUT_OF_DISTRIBUTION_ENUM_KEY)) for v in value[VALUE_SUBKEY]]
        elif metric_name == SupportedKeyWiseMetric.Detection_MeanAveragePrecision or metric_name == SupportedKeyWiseMetric.Detection_MicroPrecisionRecallF1:
            # Expects list of list of list of classes, [optionally scores], and bounding box coordinates, e.g., [[[0, 1.0, 0, 0, 10, 10]]]
            if type in SIMPLE_KEY_TYPES:
                self.key_metric_map[key]["prediction_preprocessor"] = lambda value: self._detection_preprocess_single_prediction(key, value)
                self.key_metric_map[key]["target_preprocessor"] = lambda value: self._detection_preprocess_single_target(key, value)
            elif type == JSONSchemaKeyType.Array:
                def detection_list_preprocess_prediction(value):
                    value_list = value[VALUE_SUBKEY]
                    return_pred = []
                    for v in value_list:
                        return_pred.extend(self._detection_preprocess_single_prediction(key, v))
                    return return_pred

                def detection_list_preprocess_target(value):
                    value_list = value[VALUE_SUBKEY]
                    return_gt = []
                    for v in value_list:
                        return_gt.extend(self._detection_preprocess_single_target(key, v))
                    return return_gt
                self.key_metric_map[key]["prediction_preprocessor"] = detection_list_preprocess_prediction
                self.key_metric_map[key]["target_preprocessor"] = detection_list_preprocess_target
        elif metric_name == SupportedKeyWiseMetric.Regression_MeanAbsoluteError or metric_name == SupportedKeyWiseMetric.Regression_MeanAbsoluteErrorF1Score:
            # Expects torch int or float tensor of shape (N)
            self.key_metric_map[key]["prediction_preprocessor"] = self.key_metric_map[key]["target_preprocessor"] = lambda value: float(value[VALUE_SUBKEY])



from .bleu import BleuScore
from .cider import CIDErScore
from .meteor import METEORScore
from .rougel import ROUGELScore
from .azure_openai_model_eval import AzureOpenAITextModelCategoricalScore


__all__ = ['BleuScore', 'CIDErScore', 'METEORScore', 'ROUGELScore', 'AzureOpenAITextModelCategoricalScore']



import unittest

from visionmetrics.common.utils import precision_recall_f1_scalar


class TestPrecisionRecallF1(unittest.TestCase):
    def test_perfect(self):
        report = precision_recall_f1_scalar(tp=5, fp=0, fn=0)
        self.assertEqual(report, {
            "Precision": 1.0,
            "Recall": 1.0,
            "F1": 1.0
        })

    def test_perfect_precision_partial_recall(self):
        report = precision_recall_f1_scalar(tp=5, fp=0, fn=3)
        self.assertEqual(report["Precision"], 1.0)
        self.assertEqual(report["Recall"], 0.625)
        self.assertAlmostEqual(report["F1"], 0.7692307692307692)

    def test_perfect_recall_partial_precision(self):
        report = precision_recall_f1_scalar(tp=5, fp=3, fn=0)
        self.assertEqual(report["Precision"], 0.625)
        self.assertEqual(report["Recall"], 1.0)
        self.assertAlmostEqual(report["F1"], 0.7692307692307692)

    def test_imperfect_equal_precision_recall(self):
        report = precision_recall_f1_scalar(tp=5, fp=3, fn=3)
        self.assertEqual(report, {
            "Precision": 0.625,
            "Recall": 0.625,
            "F1": 0.625
        })

    def test_imperfect_unequal_precision_recall(self):
        report = precision_recall_f1_scalar(tp=5, fp=3, fn=5)
        self.assertEqual(report, {
            "Precision": 0.625,
            "Recall": 0.5,
            "F1": 0.5555555555555556
        })

    def test_zero_recall_no_precision_samples(self):
        report = precision_recall_f1_scalar(tp=0, fp=0, fn=1)
        self.assertEqual(report, {
            "Precision": 0.0,
            "Recall": 0.0,
            "F1": 0.0
        })

    def test_zero_precision_no_recall_samples(self):
        report = precision_recall_f1_scalar(tp=0, fp=1, fn=0)
        self.assertEqual(report, {
            "Precision": 0.0,
            "Recall": 0.0,
            "F1": 0.0
        })

    def test_no_samples(self):
        report = precision_recall_f1_scalar(tp=0, fp=0, fn=0)
        self.assertEqual(report, {
            "Precision": 0.0,
            "Recall": 0.0,
            "F1": 0.0
        })


if __name__ == '__main__':
    unittest.main()



import unittest
from torch import tensor

from visionmetrics.regression.accuracy import MeanAbsoluteErrorF1Score


class TestDetectionMicroPrecisionRecallF1(unittest.TestCase):
    def test_all_perfect(self):
        evaluator = MeanAbsoluteErrorF1Score(error_threshold=0.0)
        evaluator.update(predictions=tensor([1, 2, 3, 4]), targets=tensor([1, 2, 3, 4]))
        report = evaluator.compute()
        self.assertEqual(report, {
            "Precision": 1.0,
            "Recall": 1.0,
            "F1": 1.0
        })

    def test_all_right(self):
        evaluator = MeanAbsoluteErrorF1Score(error_threshold=0.5)
        evaluator.update(predictions=tensor([1, 2, 3, 4]), targets=tensor([1.4, 2.4, 3.4, 3.6]))
        report = evaluator.compute()
        self.assertEqual(report, {
            "Precision": 1.0,
            "Recall": 1.0,
            "F1": 1.0
        })

    def test_partially_right(self):
        evaluator = MeanAbsoluteErrorF1Score(error_threshold=0.5)
        evaluator.update(predictions=tensor([1, 2, 3, 4]), targets=tensor([1, 2.6, 3.5, 4.6]))
        report = evaluator.compute()
        self.assertEqual(report, {
            "Precision": 0.5,
            "Recall": 0.5,
            "F1": 0.5
        })

    def test_all_right_equal_to_threshold(self):
        evaluator = MeanAbsoluteErrorF1Score(error_threshold=1.0)
        evaluator.update(predictions=tensor([1, 2, 3, 4]), targets=tensor([2, 3, 4, 5]))
        report = evaluator.compute()
        self.assertEqual(report, {
            "Precision": 1.0,
            "Recall": 1.0,
            "F1": 1.0
        })

    def test_all_wrong_greater_than_threshold(self):
        evaluator = MeanAbsoluteErrorF1Score(error_threshold=1.0)
        evaluator.update(predictions=tensor([1, 2, 3, 4]), targets=tensor([3, 4, 5, 6]))
        report = evaluator.compute()
        self.assertEqual(report, {
            "Precision": 0.0,
            "Recall": 0.0,
            "F1": 0.0
        })

    def test_all_perfect_matrix(self):
        evaluator = MeanAbsoluteErrorF1Score(error_threshold=0.0)
        evaluator.update(predictions=tensor([[1, 2, 3, 4], [0, 0.5, 1, 2]]),
                         targets=tensor([[1, 2, 3, 4], [0, 0.5, 1, 2]]))
        report = evaluator.compute()
        self.assertEqual(report, {
            "Precision": 1.0,
            "Recall": 1.0,
            "F1": 1.0
        })

    def test_all_right_matrix(self):
        evaluator = MeanAbsoluteErrorF1Score(error_threshold=1.0)
        evaluator.update(predictions=tensor([[1, 2, 3, 4], [0, 0.5, 1, 2]]),
                         targets=tensor([[1.1, 2.4, 3.6, 4.5], [0.1, 1, 1, 2.2]]))
        report = evaluator.compute()
        self.assertEqual(report, {
            "Precision": 1.0,
            "Recall": 1.0,
            "F1": 1.0
        })

    def test_partially_right_matrix(self):
        evaluator = MeanAbsoluteErrorF1Score(error_threshold=0.5)
        evaluator.update(predictions=tensor([[1, 2, 3, 4], [0, 0.5, 1, 2]]),
                         targets=tensor([[1.1, 2.4, 3.6, 4.6], [0.1, 1.1, 1, 2.2]]))
        report = evaluator.compute()
        self.assertEqual(report, {
            "Precision": 0.625,
            "Recall": 0.625,
            "F1": 0.625
        })

    def test_all_right_matrix_equal_to_threshold(self):
        evaluator = MeanAbsoluteErrorF1Score(error_threshold=1.0)
        evaluator.update(predictions=tensor([[1, 2, 3, 4], [0, 0.5, 1, 2]]),
                         targets=tensor([[2, 3, 4, 5], [-1, -0.5, 0, 1]]))
        report = evaluator.compute()
        self.assertEqual(report, {
            "Precision": 1.0,
            "Recall": 1.0,
            "F1": 1.0
        })

    def test_all_wrong_matrix_greater_than_threshold(self):
        evaluator = MeanAbsoluteErrorF1Score(error_threshold=1.0)
        evaluator.update(predictions=tensor([[1, 2, 3, 4], [0, 0.5, 1, 2]]),
                         targets=tensor([[2.5, 3.5, 4.5, 5.5], [-1.5, -1, -0.5, 0.5]]))
        report = evaluator.compute()
        self.assertEqual(report, {
            "Precision": 0.0,
            "Recall": 0.0,
            "F1": 0.0
        })

    def test_partially_right_nested_array(self):
        evaluator = MeanAbsoluteErrorF1Score(error_threshold=0.5)
        evaluator.update(predictions=tensor([[[1, 2, 3, 4]]]), targets=tensor([[[1, 2.6, 3.6, 4]]]))
        report = evaluator.compute()
        self.assertEqual(report, {
            "Precision": 0.5,
            "Recall": 0.5,
            "F1": 0.5
        })

    def test_predictions_not_numeric_tensor(self):
        evaluator = MeanAbsoluteErrorF1Score(error_threshold=1.0)
        self.assertRaisesRegex(ValueError,
                               r"'predictions' must be a float or integer torch.tensor.",
                               evaluator.update,
                               [1, 2, 3, 4],
                               tensor([2.5, 3.5, 4.5, 5.5])
                               )

    def test_targets_not_numeric_tensor(self):
        evaluator = MeanAbsoluteErrorF1Score(error_threshold=1.0)
        self.assertRaisesRegex(ValueError,
                               r"'targets' must be a float or integer torch.tensor.",
                               evaluator.update,
                               tensor([2.5, 3.5, 4.5, 5.5]),
                               [1, 2, 3, 4]
                               )

    def test_predictions_not_same_shape(self):
        evaluator = MeanAbsoluteErrorF1Score(error_threshold=1.0)
        self.assertRaisesRegex(ValueError,
                               r"'predictions' and 'targets' must have the same shape; got predictions of shape \[4\] and targets of shape \[1, 4\].",
                               evaluator.update,
                               tensor([2.5, 3.5, 4.5, 5.5]),
                               tensor([[1, 2, 3, 4]])
                               )

    def test_initialize_with_negative_threshold(self):
        self.assertRaisesRegex(ValueError,
                               r"'error_threshold' must be >= 0; got error_threshold=-1.0.",
                               MeanAbsoluteErrorF1Score,
                               error_threshold=-1.0)


if __name__ == '__main__':
    unittest.main()



# Import metrics directly from torchmetrics
from torchmetrics.regression import MeanAbsoluteError, MeanSquaredError

# Import custom metrics from visionmetrics
from visionmetrics.regression.accuracy import MeanAbsoluteErrorF1Score



from irisml.tasks.create_azure_openai_chat_model import OpenAITextChatModel
import json
import pathlib
import unittest

from visionmetrics.caption import BleuScore, CIDErScore, METEORScore, ROUGELScore, AzureOpenAITextModelCategoricalScore
from visionmetrics.caption.azure_openai_model_eval_base import ResultStatusType


class TestImageCaptionEvaluator(unittest.TestCase):
    predictions_file = pathlib.Path(__file__).resolve().parent / 'data' / 'image_caption_prediction.json'
    ground_truth_file = pathlib.Path(__file__).resolve().parent / 'data' / 'image_caption_gt.json'
    imcap_predictions, imcap_targets = [], []
    predictions_dict = json.loads(predictions_file.read_text())
    ground_truth_dict = json.loads(ground_truth_file.read_text())

    gts_by_id = {}
    predictions_by_id = {pred['image_id']: pred['caption'] for pred in predictions_dict}

    for gt in ground_truth_dict['annotations']:
        if not gt['image_id'] in gts_by_id:
            gts_by_id[gt['image_id']] = []
        gts_by_id[gt['image_id']].append(gt['caption'])
    for key, value in predictions_by_id.items():
        imcap_predictions.append(value)
        imcap_targets.append(gts_by_id[key])

    def test_image_caption_bleu_score_evaluator(self):
        evaluator = BleuScore()
        evaluator.update(predictions=self.imcap_predictions, targets=self.imcap_targets)
        report = evaluator.compute()
        self.assertAlmostEqual(report["Bleu_1"], 0.783228681385441)
        self.assertAlmostEqual(report["Bleu_2"], 0.6226378540059051)
        self.assertAlmostEqual(report["Bleu_3"], 0.47542636331846966)
        self.assertAlmostEqual(report["Bleu_4"], 0.3573567238999926)

    def test_image_caption_meteor_score_evaluator(self):
        evaluator = METEORScore()
        evaluator.update(predictions=self.imcap_predictions, targets=self.imcap_targets)
        report = evaluator.compute()
        self.assertAlmostEqual(report["METEOR"], 0.2878681068021112)

    def test_image_caption_rouge_l_score_evaluator(self):
        evaluator = ROUGELScore()
        evaluator.update(predictions=self.imcap_predictions, targets=self.imcap_targets)
        report = evaluator.compute()
        self.assertAlmostEqual(report["ROUGE_L"], 0.5774238052522583)

    def test_image_caption_cider_score_evaluator(self):
        evaluator = CIDErScore()
        evaluator.update(predictions=self.imcap_predictions, targets=self.imcap_targets)
        report = evaluator.compute()
        self.assertAlmostEqual(report["CIDEr"], 1.2346054374217474)


class TestAzureOpenAITextModelCategoricalEvaluator(unittest.TestCase):
    tp_exact_match_testcase = {
        "predictions": ["This is an image of a dog."],
        "targets": [["This is an image of a dog."]]
    }
    tp_semantic_match_testcase = {
        "predictions": ["In this image, there is a dog."],
        "targets": [["This is an image of a dog."]]
    }
    tn_testcase = {
        "predictions": [""],
        "targets": [""]
    }
    fn_testcase = {
        "predictions": [""],
        "targets": [["This is an image of a dog."]]
    }
    fp_gt_null_testcase = {
        "predictions": ["In this image, there is a dog."],
        "targets": [[""]]
    }
    fp_gt_not_null_testcase = {
        "predictions": ["In this image, there is a dog."],
        "targets": [["There is a cat on a bench."]]
    }

    # Since AverageScore and ScoreParseFailures are not deterministic, omitting these from the testcase
    expected_report = {
        # Summary statistics
        "Precision": 0.5,
        "Recall": 0.5,
        "F1": 0.5,
        "Accuracy": 0.5,
        # Raw statistic counts
        "TP": 2,
        "TN": 1,
        "FN": 1,
        "FP_GTNull": 1,
        "FP_GTNotNull": 1
    }

    def test_azure_openai_text_model_categorical_evaluator(self):
        # Update with each test case and check intermediate states for correctness, followed by overall metrics.
        evaluator = AzureOpenAITextModelCategoricalScore(endpoint="https://endpoint-name.openai.azure.com/", deployment_name="gpt-4o")

        with unittest.mock.patch.object(OpenAITextChatModel, "forward", return_value=["1.0"]):
            evaluator.update(predictions=self.tp_exact_match_testcase["predictions"], targets=self.tp_exact_match_testcase["targets"])
            self.assertEqual(evaluator.raw_scores[-1], "1.0")
            self.assertEqual(evaluator.scores[-1], 1.0)
            self.assertEqual(evaluator.score_parse_failures.item(), 0)
            self.assertEqual(evaluator.result_status_types[-1], ResultStatusType.TruePositive)

        with unittest.mock.patch.object(OpenAITextChatModel, "forward", return_value=["0.9"]):
            evaluator.update(predictions=self.tp_semantic_match_testcase["predictions"], targets=self.tp_semantic_match_testcase["targets"])
            self.assertGreater(evaluator.scores[-1], evaluator.positive_threshold)
            self.assertEqual(evaluator.score_parse_failures.item(), 0)
            self.assertEqual(evaluator.result_status_types[-1], ResultStatusType.TruePositive)

        with unittest.mock.patch.object(OpenAITextChatModel, "forward", return_value=["1.0"]):
            evaluator.update(predictions=self.tn_testcase["predictions"], targets=self.tn_testcase["targets"])
            self.assertEqual(evaluator.raw_scores[-1], "1.0")
            self.assertEqual(evaluator.scores[-1], 1.0)
            self.assertEqual(evaluator.score_parse_failures.item(), 0)
            self.assertEqual(evaluator.result_status_types[-1], ResultStatusType.TrueNegative)

        with unittest.mock.patch.object(OpenAITextChatModel, "forward", return_value=["0.0"]):
            evaluator.update(predictions=self.fn_testcase["predictions"], targets=self.fn_testcase["targets"])
            self.assertLess(evaluator.scores[-1], evaluator.positive_threshold)
            self.assertEqual(evaluator.result_status_types[-1], ResultStatusType.FalseNegative)

        with unittest.mock.patch.object(OpenAITextChatModel, "forward", return_value=["0.0"]):
            evaluator.update(predictions=self.fp_gt_null_testcase["predictions"], targets=self.fp_gt_null_testcase["targets"])
            self.assertLess(evaluator.scores[-1], evaluator.positive_threshold)
            self.assertEqual(evaluator.result_status_types[-1], ResultStatusType.FalsePositiveGtNull)

        with unittest.mock.patch.object(OpenAITextChatModel, "forward", return_value=["0.0"]):
            evaluator.update(predictions=self.fp_gt_not_null_testcase["predictions"], targets=self.fp_gt_not_null_testcase["targets"])
            self.assertLess(evaluator.scores[-1], evaluator.positive_threshold)
            self.assertEqual(evaluator.result_status_types[-1], ResultStatusType.FalsePositiveGtNotNull)

        report = evaluator.compute()
        self.assertAlmostEqual(report["Precision"], self.expected_report["Precision"])
        self.assertAlmostEqual(report["Recall"], self.expected_report["Recall"])
        self.assertAlmostEqual(report["F1"], self.expected_report["F1"])
        self.assertAlmostEqual(report["Accuracy"], self.expected_report["Accuracy"])
        self.assertEqual(report["TP"], self.expected_report["TP"])
        self.assertEqual(report["TN"], self.expected_report["TN"])
        self.assertEqual(report["FN"], self.expected_report["FN"])
        self.assertEqual(report["FP_GTNull"], self.expected_report["FP_GTNull"])
        self.assertEqual(report["FP_GTNotNull"], self.expected_report["FP_GTNotNull"])



from collections import Counter
import torch
import torchmetrics

from visionmetrics.classification.common import MulticlassMixin, MultilabelMixin
from visionmetrics.common.utils import precision_recall_f1_scalar


class MulticlassPrecision(MulticlassMixin, torchmetrics.classification.MulticlassPrecision):
    pass


class MulticlassRecall(MulticlassMixin, torchmetrics.classification.MulticlassRecall):
    pass


class MulticlassF1Score(MulticlassMixin, torchmetrics.classification.MulticlassF1Score):
    pass


class MultilabelPrecision(MultilabelMixin, torchmetrics.classification.MultilabelPrecision):
    pass


class MultilabelRecall(MultilabelMixin, torchmetrics.classification.MultilabelRecall):
    pass


class MultilabelF1Score(MultilabelMixin, torchmetrics.classification.MultilabelF1Score):
    pass


class MultilabelF1ScoreWithDuplicates(torchmetrics.Metric):
    """
    Calculates and returns a variant of multilabel F1 Score where, instead of considering sets (no duplicate values) of tags, lists of tags with potential duplicates are permitted.
    """
    def __init__(self):
        super().__init__()

        self.add_state("tp", default=torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("fp", default=torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("fn", default=torch.tensor(0.0), dist_reduce_fx="sum")

    def update(self, predictions: list, targets: list):
        if len(predictions) != len(targets):
            raise ValueError(f"Number of predictions and targets should be the same, but got predictions of length {len(predictions)} and targets of length {len(targets)}.")
        for prediction, target in zip(predictions, targets):
            pred_counts = Counter(prediction)
            target_counts = Counter(target)
            self.tp += sum([min(pred_counts[k], target_counts[k]) for k in pred_counts if k in target_counts])
            self.fp += sum([max(pred_counts[k] - target_counts[k], 0) for k in pred_counts if k in target_counts] + [pred_counts[k] for k in pred_counts if k not in target_counts])
            self.fn += sum([max(target_counts[k] - pred_counts[k], 0) for k in target_counts if k in pred_counts] + [target_counts[k] for k in target_counts if k not in pred_counts])

    def compute(self):
        tp = self.tp.item()
        fp = self.fp.item()
        fn = self.fn.item()
        return precision_recall_f1_scalar(tp=tp, fp=fp, fn=fn)



import torch
from torchmetrics import Metric

from visionmetrics.common.utils import precision_recall_f1_scalar


class MeanAbsoluteErrorF1Score(Metric):
    """
    Calculates and returns a discretized version of the MeanAbsoluteError metric.
    Args:
        error_threshold: float indicating the exclusive threshold below which the mean absolute error is considered a true positive.
    """
    def __init__(self, error_threshold=0.0):
        super().__init__()
        if error_threshold < 0:
            raise ValueError(f"'error_threshold' must be >= 0; got error_threshold={error_threshold}.")
        self.error_threshold = error_threshold

        self.add_state("tp", default=torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("fp", default=torch.tensor(0.0), dist_reduce_fx="sum")
        self.add_state("fn", default=torch.tensor(0.0), dist_reduce_fx="sum")

    def update(self, predictions, targets):
        if not torch.is_tensor(predictions):
            raise ValueError("'predictions' must be a float or integer torch.tensor.")
        if not torch.is_tensor(targets):
            raise ValueError("'targets' must be a float or integer torch.tensor.")
        if predictions.shape != targets.shape:
            raise ValueError(f"'predictions' and 'targets' must have the same shape; got predictions of shape {list(predictions.shape)} and targets of shape {list(targets.shape)}.")
        n_obs = targets.numel()
        tp = torch.sum(torch.where(torch.abs(predictions - targets) <= self.error_threshold, 1, 0))
        fp = n_obs - tp
        fn = n_obs - tp
        self.tp += tp
        self.fp += fp
        self.fn += fn

    def compute(self):
        # Note that we currently do not distinguish false positives vs false negatives (as in, false positive and false negative are the same),
        # we effectively have that precision == recall == F1 == accuracy.
        tp = self.tp.item()
        fp = self.fp.item()
        fn = self.fn.item()
        return precision_recall_f1_scalar(tp=tp, fp=fp, fn=fn)



import torch
from torchmetrics import Metric


class Recall(Metric):
    """
    Compute Recall@k for object grounding task.
    """

    def __init__(self, iou_thresh=0.5, k=1):
        super().__init__()
        if not (0 <= iou_thresh <= 1):
            raise ValueError(f"iou_thresh must be in [0, 1], got {iou_thresh}.")

        if not isinstance(k, int) or not (k >= 1):
            raise ValueError(f"k must be a postive integer, got {k}.")

        self.iou_thresh = iou_thresh
        self.topk = k
        self.add_state("predictions", default=[], dist_reduce_fx="cat")
        self.add_state("targets", default=[], dist_reduce_fx="cat")

    def _box_area(self, boxes):
        assert boxes.ndim == 2 and boxes.shape[-1] == 4
        return (boxes[:, 2] - boxes[:, 0]) * (boxes[:, 3] - boxes[:, 1])

    def _box_iou(self, boxes1, boxes2):
        """
        Compute intersection-over-union of input boxes. Both sets of boxes are expected to be in (top, left, bottom, right) format.

        Args:
            boxes1 (Tensor[N, 4])
            boxes2 (Tensor[M, 4])

        Returns:
            iou (Tensor[N, M]): the NxM matrix containing the pairwise IoU values for every element in boxes1 and boxes2
        """

        area1 = self._box_area(boxes1)
        area2 = self._box_area(boxes2)
        lt = torch.maximum(boxes1[:, None, :2], boxes2[:, :2])
        rb = torch.minimum(boxes1[:, None, 2:], boxes2[:, 2:])
        wh = (rb - lt).clip(min=0)
        inter = wh[:, :, 0] * wh[:, :, 1]
        union = area1[:, None] + area2 - inter
        iou = inter / union
        return iou

    def update(self, predictions, targets):
        """
        Args:
            predictions: list of tuple predictions (pred_phrases, pred_bboxes) for image object grounding task:
                         [([pred_phrase1, pred_phrase2, ...], [[pred_bbox, pred_bbox, ...], [pred_bbox, pred_bbox, ...], ...]), ...], type: list of tuple.
                         "pred_bboxes" is a list of list of bbox (top, left, bottom, right) in absolute scale.
            targets: list of tuple ground truth (target_phrases, target_bboxes) for image object grounding task:
                     [([target_phrase1, target_phrase2, ...], [[target_bbox, target_bbox, ...], [target_bbox, target_bbox, ...], ...]), ...], type: list of tuple.
                     "target_bboxes" is a list of list of bbox (top, left, bottom, right) in absolute scale.
        """
        if len(predictions) != len(targets):
            raise ValueError(f"Number of predictions and targets must be equal, got {len(predictions)} and {len(targets)}.")

        for i, (pred, target) in enumerate(zip(predictions, targets)):
            pred_phrases, pred_bboxes = pred
            target_phrases, target_bboxes = target
            if len(pred_phrases) != len(pred_bboxes):
                raise ValueError(f"Number of predicted phrases and predicted bboxes must be equal, got {len(pred_phrases)} and {len(pred_bboxes)}, index: {i}.")

            if len(target_phrases) != len(target_bboxes):
                raise ValueError(f"Number of target phrases and target bboxes must be equal, got {len(target_phrases)} and {len(target_bboxes)}, index: {i}.")

        self.targets += targets
        self.predictions += predictions

    def compute(self, **kwargs):
        """
        Returns:
            recall@k: top-k recall score
        """
        phrase_num = 0
        true_positive = 0

        for pred, target in zip(self.predictions, self.targets):
            pred_phrases, pred_bboxes = pred
            target_phrases, target_bboxes = target
            for target_phrase, target_bbox in zip(target_phrases, target_bboxes):
                phrase_num += 1
                pred_phrases = [phrase.lower() for phrase in pred_phrases]
                target_phrase = target_phrase.lower()
                if target_phrase in pred_phrases:
                    # Only consider the first matched prediction, need to be adjusted for multiple prediction
                    cur_boxes = pred_bboxes[pred_phrases.index(target_phrase)]
                    if len(cur_boxes) > 0 and len(target_bbox) > 0:
                        ious = self._box_iou(torch.tensor(cur_boxes), torch.tensor(target_bbox))
                        if ious[:self.topk].max() >= self.iou_thresh:
                            true_positive += 1
        return {f'recall@{self.topk}': true_positive / phrase_num}



import unittest
import torch


from visionmetrics.detection import (ClassAgnosticAveragePrecision,
                                     DetectionConfusionMatrix,
                                     MeanAveragePrecision,
                                     DetectionMicroPrecisionRecallF1)


class TestMeanAveragePrecision(unittest.TestCase):
    def test_perfect_one_image_absolute_coordinates(self):

        PREDICTIONS = [[[0, 1.0, 0, 0, 10, 10],
                        [1, 1.0, 5, 5, 10, 10],
                        [2, 1.0, 1, 1, 5, 5]]]

        TARGETS = [[[0, 0, 0, 10, 10],
                    [1, 5, 5, 10, 10],
                    [2, 1, 1, 5, 5]]]

        metric = MeanAveragePrecision()
        metric.update(PREDICTIONS, TARGETS)
        result = metric.compute()
        self.assertEqual(result['map_50'], 1.0)

    def test_perfect_one_image_relative_coordinates(self):

        PREDICTIONS = [[[0, 1.0, 0, 0, 1, 1],
                        [1, 1.0, 0.5, 0.5, 1, 1],
                        [2, 1.0, 0.1, 0.1, 0.5, 0.5]]]

        TARGETS = [[[0, 0, 0, 1, 1],
                    [1, 0.5, 0.5, 1, 1],
                    [2, 0.1, 0.1, 0.5, 0.5]]]

        metric = MeanAveragePrecision()
        metric.update(PREDICTIONS, TARGETS)
        result = metric.compute()
        self.assertEqual(result['map_50'], 1.0)

    def test_wrong_one_image(self):
        PREDICTIONS = [[[0, 1.0, 0, 0, 1, 1],
                        [0, 1.0, 0.5, 0.5, 1, 1],
                        [1, 1.0, 0.5, 0.5, 1, 1]]]

        TARGETS = [[[0, 0, 0, 1, 1],
                    [1, 0.5, 0.5, 1, 1]]]

        metric = MeanAveragePrecision()
        metric.update(PREDICTIONS, TARGETS)
        result = metric.compute()
        self.assertEqual(result['map_50'], 1.0)

    def test_perfect_two_images(self):
        PREDICTIONS = [[[0, 1.0, 0, 0, 1, 1],
                        [1, 1.0, 0.5, 0.5, 1, 1]],
                       [[2, 1.0, 0.1, 0.1, 0.5, 0.5]]]

        TARGETS = [[[0, 0, 0, 1, 1],
                    [1, 0.5, 0.5, 1, 1]],
                   [[2, 0.1, 0.1, 0.5, 0.5]]]

        metric = MeanAveragePrecision()
        metric.update(PREDICTIONS, TARGETS)
        result = metric.compute()
        self.assertEqual(result['map_50'], 1.0)

    def test_two_batches(self):

        PREDICTIONS = [[[0, 1.0, 0, 0, 1, 1],
                        [1, 1.0, 0.5, 0.5, 1, 1]],
                       [[2, 1.0, 0.1, 0.1, 0.5, 0.5]]]

        TARGETS = [[[0, 0, 0, 1, 1],
                    [1, 0.5, 0.5, 1, 1]],
                   [[2, 0.1, 0.1, 0.5, 0.5]]]

        metric = MeanAveragePrecision()
        metric.update(PREDICTIONS, TARGETS)

        PREDICTIONS = [[[0, 1.0, 0.9, 0.9, 1, 1],  # Wrong
                        [1, 1.0, 0.5, 0.5, 1, 1]],
                       [[2, 1.0, 0.1, 0.1, 0.5, 0.5]]]

        TARGETS = [[[0, 0, 0, 1, 1],
                    [1, 0.5, 0.5, 1, 1]],
                   [[2, 0.1, 0.1, 0.5, 0.5]]]

        metric.update(PREDICTIONS, TARGETS)

        result = metric.compute()
        self.assertAlmostEqual(result['map_50'].item(), 0.834983, places=5)

    def test_iou_thresholds(self):
        metric = MeanAveragePrecision(iou_thresholds=[0.5])
        PREDICTIONS = [[[0, 1.0, 0.5, 0.5, 1, 1],
                        [1, 1.0, 0.5, 0.5, 1, 1]]]

        TARGETS = [[[0, 0, 0, 1, 1],
                    [1, 0.5, 0.5, 1, 1]]]

        metric.update(PREDICTIONS, TARGETS)
        result = metric.compute()
        self.assertAlmostEqual(result['map_50'].item(), 0.5, places=5)

        metric = MeanAveragePrecision(iou_thresholds=[0.2])
        PREDICTIONS = [[[0, 1.0, 0.5, 0.5, 1, 1],
                        [1, 1.0, 0.5, 0.5, 1, 1]]]

        TARGETS = [[[0, 0, 0, 1, 1],
                    [1, 0.5, 0.5, 1, 1]]]

        metric.update(PREDICTIONS, TARGETS)
        result = metric.compute()
        self.assertAlmostEqual(result['map'].item(), 1.0, places=5)

    def test_no_predictions(self):
        PREDICTIONS = [[]]
        TARGETS = [[[0, 0, 0, 1, 1],
                    [1, 0.5, 0.5, 1, 1],
                    [2, 0.1, 0.1, 0.5, 0.5]]]

        metric = MeanAveragePrecision()
        metric.update(PREDICTIONS, TARGETS)
        result = metric.compute()
        self.assertAlmostEqual(result['map_50'].item(), 0.0, places=5)

    def test_no_targets(self):
        PREDICTIONS = [[[0, 1.0, 0, 0, 1, 1],
                        [1, 1.0, 0.5, 0.5, 1, 1],
                        [2, 1.0, 0.1, 0.1, 0.5, 0.5]]]
        TARGETS = [[]]

        metric = MeanAveragePrecision(iou_thresholds=[0.5])
        metric.update(PREDICTIONS, TARGETS)
        result = metric.compute()
        self.assertAlmostEqual(result['map_50'].item(), -1, places=5)

    def test_cat_id_remap(self):
        PREDICTIONS = [[(0, 1.0, 0, 0, 1, 1),
                        [1, 1.0, 0.5, 0.5, 1, 1],
                        [2, 1.0, 0.1, 0.1, 0.5, 0.5]]]

        TARGETS = [[[0, 0, 0, 1, 1],
                    [0, 0.5, 0.5, 1, 1],
                    [2, 0.1, 0.1, 0.5, 0.5]]]

        metric = MeanAveragePrecision(iou_thresholds=[0.5])
        metric.update(PREDICTIONS, TARGETS)
        result = metric.compute()

        PREDICTIONS_REMAP_CAT_ID = [[[0, 1.0, 0, 0, 1, 1],
                                     [2, 1.0, 0.5, 0.5, 1, 1],
                                     [1, 1.0, 0.1, 0.1, 0.5, 0.5]]]

        TARGETS_REMAP_CAT_ID = [[[0, 0, 0, 1, 1],
                                 [0, 0.5, 0.5, 1, 1],
                                 [1, 0.1, 0.1, 0.5, 0.5]]]

        metric = MeanAveragePrecision(iou_thresholds=[0.5])
        metric.update(PREDICTIONS_REMAP_CAT_ID, TARGETS_REMAP_CAT_ID)
        result_remap_cat_id = metric.compute()

        for k in result.keys():
            self.assertTrue(torch.allclose(result[k], result_remap_cat_id[k]))

    def test_map_small_medium_large(self):
        PREDICTIONS = [[[0, 1.0, 0, 0, 32, 32],  # <= 32**2
                        [1, 1.0, 0, 0, 96, 96],  # >= 32**2, <= 96**2
                        [2, 1.0, 0, 0, 100, 100]]]  # > 96**2

        TARGETS = [[[0, 0, 0, 32, 32],
                    [1, 5, 5, 96, 96],
                    [2, 1, 1, 100, 100]]]

        metric = MeanAveragePrecision(coords='absolute', iou_thresholds=[0.5])
        metric.update(PREDICTIONS, TARGETS)
        result = metric.compute()

        self.assertAlmostEqual(result['map_small'].item(), 1.0, places=5)
        self.assertAlmostEqual(result['map_medium'].item(), 1.0, places=5)
        self.assertAlmostEqual(result['map_large'].item(), 1.0, places=5)


class TestClassAgnosticAveragePrecision(unittest.TestCase):
    def test_class_agnostic_ap_correct(self):
        PREDICTIONS = [[[0, 1.0, 0, 0, 1, 1]]]
        TARGETS = [[[1, 0, 0, 1, 1]]]

        metric = ClassAgnosticAveragePrecision(iou_thresholds=[0.5], class_metrics=False)
        metric.update(PREDICTIONS, TARGETS)
        result = metric.compute()
        self.assertEqual(result['map_50'], 1.0)
        self.assertEqual(result['classes'], -1)
        self.assertEqual(result['map_per_class'], -1)

        metric = ClassAgnosticAveragePrecision(iou_thresholds=[0.5], class_metrics=True)
        metric.update(PREDICTIONS, TARGETS)
        result = metric.compute()
        self.assertEqual(result['classes'], -1)
        self.assertEqual(result['map_per_class'], 1)

    def test_class_agnostic_ap_incorrect(self):
        PREDICTIONS = [[[0, 1.0, 0, 0, 1, 1]]]
        TARGETS = [[[1, 0, 0, 0.1, 0.1]]]
        metric = ClassAgnosticAveragePrecision(iou_thresholds=[0.5])
        metric.update(PREDICTIONS, TARGETS)
        result = metric.compute()
        self.assertEqual(result['map_50'], 0.0)
        self.assertEqual(result['classes'], -1)
        self.assertEqual(result['map_per_class'], -1)

    def test_two_images(self):
        PREDICTIONS = [[[0, 1.0, 0, 0, 1, 1],
                        [1, 1.0, 0.5, 0.5, 1, 1]],
                       [[2, 1.0, 0.1, 0.1, 0.5, 0.5]]]

        TARGETS = [[[1, 0, 0, 1, 1],
                    [2, 0.5, 0.5, 1, 1]],
                   [[1, 0.1, 0.1, 0.5, 0.5]]]

        metric = ClassAgnosticAveragePrecision(iou_thresholds=[0.5])
        metric.update(PREDICTIONS, TARGETS)
        result = metric.compute()
        self.assertEqual(result['map_50'], 1.0)
        self.assertEqual(result['classes'], -1)


class TestDetectionConfusionMatrix(unittest.TestCase):
    def test_true_positive(self):
        predictions = [[[0, 1.0, 0, 0, 1, 1]], [[1, 1.0, 0, 0, 1, 1]]]
        targets = [[[0, 0, 0, 1, 1]], [[1, 0, 0, 1, 1]]]

        metric = DetectionConfusionMatrix(iou_threshold=0.5)
        metric.update(predictions, targets)
        result = metric.compute()

        self.assertEqual(result['TP'], 2)
        self.assertEqual(result['FP'], 0)
        self.assertEqual(result['FN'], 0)
        self.assertEqual(result['FP_due_to_wrong_class'], 0)
        self.assertEqual(result['FP_due_to_low_iou_correct_class'], 0)
        self.assertEqual(result['FP_due_to_low_iou_wrong_class'], 0)
        self.assertEqual(result['FP_due_to_extra_pred_boxes'], 0)

    def test_false_positive_wrong_class(self):
        predictions = [[[0, 1.0, 0, 0, 1, 1]], [[1, 1.0, 0, 0, 1, 1]]]
        targets = [[[0, 0, 0, 1, 1]], [[0, 0, 0, 1, 1]]]

        metric = DetectionConfusionMatrix(iou_threshold=0.5)
        metric.update(predictions, targets)
        result = metric.compute()

        self.assertEqual(result['TP'], 1)
        self.assertEqual(result['FP'], 1)
        self.assertEqual(result['FN'], 1)
        self.assertEqual(result['FP_due_to_wrong_class'], 1)
        self.assertEqual(result['FP_due_to_low_iou_correct_class'], 0)
        self.assertEqual(result['FP_due_to_low_iou_wrong_class'], 0)
        self.assertEqual(result['FP_due_to_extra_pred_boxes'], 0)

    def test_false_positive_low_iou_correct_class(self):
        predictions = [[[0, 1.0, 0, 0, 1, 1]], [[1, 1.0, 0, 0, 0.5, 0.5]]]
        targets = [[[0, 0, 0, 1, 1]], [[1, 0, 0, 1, 1]]]

        metric = DetectionConfusionMatrix(iou_threshold=0.5)
        metric.update(predictions, targets)
        result = metric.compute()

        self.assertEqual(result['TP'], 1)
        self.assertEqual(result['FP'], 1)
        self.assertEqual(result['FN'], 1)
        self.assertEqual(result['FP_due_to_wrong_class'], 0)
        self.assertEqual(result['FP_due_to_low_iou_correct_class'], 1)
        self.assertEqual(result['FP_due_to_low_iou_wrong_class'], 0)
        self.assertEqual(result['FP_due_to_extra_pred_boxes'], 0)

    def test_false_positive_low_iou_wrong_class(self):
        predictions = [[[0, 1.0, 0, 0, 1, 1]], [[1, 1.0, 0, 0, 0.5, 0.5]]]
        targets = [[[0, 0, 0, 1, 1]], [[0, 0, 0, 1, 1]]]  # wrong class

        metric = DetectionConfusionMatrix(iou_threshold=0.5)
        metric.update(predictions, targets)
        result = metric.compute()

        self.assertEqual(result['TP'], 1)
        self.assertEqual(result['FP'], 1)
        self.assertEqual(result['FN'], 1)
        self.assertEqual(result['FP_due_to_wrong_class'], 0)
        self.assertEqual(result['FP_due_to_low_iou_correct_class'], 0)
        self.assertEqual(result['FP_due_to_low_iou_wrong_class'], 1)
        self.assertEqual(result['FP_due_to_extra_pred_boxes'], 0)

    def test_false_positive_no_overlap_correct_class(self):
        predictions = [[[1, 1.0, 0, 0, 0.1, 0.1]]]
        targets = [[[1, 0.2, 0.2, 1, 1]]]

        metric = DetectionConfusionMatrix(iou_threshold=0.5)
        metric.update(predictions, targets)
        result = metric.compute()

        self.assertEqual(result['TP'], 0)
        self.assertEqual(result['FP'], 1)
        self.assertEqual(result['FN'], 1)
        self.assertEqual(result['FP_due_to_wrong_class'], 0)
        self.assertEqual(result['FP_due_to_low_iou_correct_class'], 1)
        self.assertEqual(result['FP_due_to_low_iou_wrong_class'], 0)
        self.assertEqual(result['FP_due_to_extra_pred_boxes'], 0)

    def test_false_positive_no_overlap_wrong_class(self):
        predictions = [[[0, 1.0, 0, 0, 10, 10]]]
        targets = [[[1, 90, 90, 100, 100]]]

        metric = DetectionConfusionMatrix(iou_threshold=0.5)
        metric.update(predictions, targets)
        result = metric.compute()

        self.assertEqual(result['TP'], 0)
        self.assertEqual(result['FP'], 1)
        self.assertEqual(result['FN'], 1)
        self.assertEqual(result['FP_due_to_wrong_class'], 0)
        self.assertEqual(result['FP_due_to_low_iou_correct_class'], 0)
        self.assertEqual(result['FP_due_to_low_iou_wrong_class'], 1)
        self.assertEqual(result['FP_due_to_extra_pred_boxes'], 0)

    def test_false_positive_extra_pred_boxes(self):
        predictions = [[[0, 1.0, 0, 0, 1, 1], [1, 1.0, 0, 0, 1, 1]]]
        targets = [[[0, 0, 0, 1, 1]]]

        metric = DetectionConfusionMatrix(iou_threshold=0.5)
        metric.update(predictions, targets)
        result = metric.compute()

        self.assertEqual(result['TP'], 1)
        self.assertEqual(result['FP'], 1)
        self.assertEqual(result['FN'], 0)
        self.assertEqual(result['FP_due_to_wrong_class'], 0)
        self.assertEqual(result['FP_due_to_low_iou_correct_class'], 0)
        self.assertEqual(result['FP_due_to_low_iou_wrong_class'], 0)
        self.assertEqual(result['FP_due_to_extra_pred_boxes'], 1)

    def test_false_negative(self):
        predictions = [[[0, 1.0, 0, 0, 1, 1]]]
        targets = [[[0, 0, 0, 1, 1], [1, 0, 0, 1, 1]]]

        metric = DetectionConfusionMatrix(iou_threshold=0.5)
        metric.update(predictions, targets)
        result = metric.compute()

        self.assertEqual(result['TP'], 1)
        self.assertEqual(result['FP'], 0)
        self.assertEqual(result['FN'], 1)
        self.assertEqual(result['FP_due_to_wrong_class'], 0)
        self.assertEqual(result['FP_due_to_low_iou_correct_class'], 0)
        self.assertEqual(result['FP_due_to_low_iou_wrong_class'], 0)
        self.assertEqual(result['FP_due_to_extra_pred_boxes'], 0)

    def test_batch_update(self):
        predictions = [[[0, 1.0, 0, 0, 1, 1]]]
        targets = [[[0, 0, 0, 1, 1]]]

        metric = DetectionConfusionMatrix(iou_threshold=0.5)
        metric.update(predictions, targets)
        result = metric.compute()

        self.assertEqual(result['TP'], 1)
        self.assertEqual(result['FP'], 0)
        self.assertEqual(result['FN'], 0)
        self.assertEqual(result['FP_due_to_wrong_class'], 0)
        self.assertEqual(result['FP_due_to_low_iou_correct_class'], 0)
        self.assertEqual(result['FP_due_to_low_iou_wrong_class'], 0)
        self.assertEqual(result['FP_due_to_extra_pred_boxes'], 0)

        predictions = [[[0, 1.0, 0, 0, 1, 1]]]
        targets = [[[1, 0, 0, 1, 1]]]
        metric.update(predictions, targets)
        result = metric.compute()

        self.assertEqual(result['TP'], 1)
        self.assertEqual(result['FP'], 1)
        self.assertEqual(result['FN'], 1)
        self.assertEqual(result['FP_due_to_wrong_class'], 1)
        self.assertEqual(result['FP_due_to_low_iou_correct_class'], 0)
        self.assertEqual(result['FP_due_to_low_iou_wrong_class'], 0)
        self.assertEqual(result['FP_due_to_extra_pred_boxes'], 0)

    def test_empty_predictions(self):
        predictions = [[[]]]
        targets = [[[0, 0, 0, 1, 1]]]

        metric = DetectionConfusionMatrix(iou_threshold=0.5)
        metric.update(predictions, targets)
        result = metric.compute()

        self.assertEqual(result['TP'], 0)
        self.assertEqual(result['FP'], 0)
        self.assertEqual(result['FN'], 1)
        self.assertEqual(result['FP_due_to_wrong_class'], 0)
        self.assertEqual(result['FP_due_to_low_iou_correct_class'], 0)
        self.assertEqual(result['FP_due_to_low_iou_wrong_class'], 0)
        self.assertEqual(result['FP_due_to_extra_pred_boxes'], 0)

    def test_empty_targets(self):
        predictions = [[[0, 1.0, 0, 0, 1, 1]]]
        targets = [[]]

        metric = DetectionConfusionMatrix(iou_threshold=0.5)
        metric.update(predictions, targets)
        result = metric.compute()

        self.assertEqual(result['TP'], 0)
        self.assertEqual(result['FP'], 1)
        self.assertEqual(result['FN'], 0)
        self.assertEqual(result['FP_due_to_wrong_class'], 0)
        self.assertEqual(result['FP_due_to_low_iou_correct_class'], 0)
        self.assertEqual(result['FP_due_to_low_iou_wrong_class'], 0)
        self.assertEqual(result['FP_due_to_extra_pred_boxes'], 1)

    def test_two_images(self):
        predictions = [[[0, 1.0, 0, 0, 1, 1]], [[1, 1.0, 0, 0, 1, 1]]]
        targets = [[[0, 0, 0, 1, 1]], [[1, 0, 0, 1, 1]]]

        metric = DetectionConfusionMatrix(iou_threshold=0.5)
        metric.update(predictions, targets)
        result = metric.compute()

        self.assertEqual(result['TP'], 2)
        self.assertEqual(result['FP'], 0)
        self.assertEqual(result['FN'], 0)
        self.assertEqual(result['FP_due_to_wrong_class'], 0)
        self.assertEqual(result['FP_due_to_low_iou_correct_class'], 0)
        self.assertEqual(result['FP_due_to_low_iou_wrong_class'], 0)
        self.assertEqual(result['FP_due_to_extra_pred_boxes'], 0)

    def test_two_images_one_wrong_class_one_low_iou(self):
        predictions = [[[0, 1.0, 0, 0, 1, 1]], [[1, 1.0, 0, 0, 0.5, 0.5]]]
        targets = [[[1, 0, 0, 1, 1]], [[1, 0, 0, 1, 1]]]

        metric = DetectionConfusionMatrix(iou_threshold=0.5)
        metric.update(predictions, targets)
        result = metric.compute()

        self.assertEqual(result['TP'], 0)
        self.assertEqual(result['FP'], 2)
        self.assertEqual(result['FN'], 2)
        self.assertEqual(result['FP_due_to_wrong_class'], 1)
        self.assertEqual(result['FP_due_to_low_iou_correct_class'], 1)
        self.assertEqual(result['FP_due_to_low_iou_wrong_class'], 0)
        self.assertEqual(result['FP_due_to_extra_pred_boxes'], 0)

    def test_two_images_extra_preds(self):
        predictions = [[[0, 1.0, 0, 0, 1, 1], [1, 1.0, 0, 0, 1, 1]], [[1, 1.0, 0, 0, 1, 1]]]
        targets = [[[0, 0, 0, 1, 1]], [[1, 0, 0, 1, 1]]]

        metric = DetectionConfusionMatrix(iou_threshold=0.5)
        metric.update(predictions, targets)
        result = metric.compute()

        self.assertEqual(result['TP'], 2)
        self.assertEqual(result['FP'], 1)
        self.assertEqual(result['FN'], 0)
        self.assertEqual(result['FP_due_to_wrong_class'], 0)
        self.assertEqual(result['FP_due_to_low_iou_correct_class'], 0)
        self.assertEqual(result['FP_due_to_low_iou_wrong_class'], 0)
        self.assertEqual(result['FP_due_to_extra_pred_boxes'], 1)

    def test_two_images_extra_targets(self):
        predictions = [[[0, 1.0, 0, 0, 1, 1]], [[1, 1.0, 0, 0, 1, 1]]]
        targets = [[[0, 0, 0, 1, 1], [1, 0, 0, 1, 1]], [[1, 0, 0, 1, 1]]]

        metric = DetectionConfusionMatrix(iou_threshold=0.5)
        metric.update(predictions, targets)
        result = metric.compute()

        self.assertEqual(result['TP'], 2)
        self.assertEqual(result['FP'], 0)
        self.assertEqual(result['FN'], 1)
        self.assertEqual(result['FP_due_to_wrong_class'], 0)
        self.assertEqual(result['FP_due_to_low_iou_correct_class'], 0)
        self.assertEqual(result['FP_due_to_low_iou_wrong_class'], 0)
        self.assertEqual(result['FP_due_to_extra_pred_boxes'], 0)

    def test_two_images_zero_area_box(self):
        predictions = [[[1, 1.0, 0, 0, 100, 100]], [[1, 1.0, 20, 20, 20, 20]]]
        targets = [[[1, 0, 0, 100, 100]], [[1, 20, 20, 20, 20]]]

        metric = DetectionConfusionMatrix(iou_threshold=0.5)
        metric.update(predictions, targets)
        result = metric.compute()

        self.assertEqual(result['TP'], 2)
        self.assertEqual(result['FP'], 0)
        self.assertEqual(result['FN'], 0)
        self.assertEqual(result['FP_due_to_wrong_class'], 0)
        self.assertEqual(result['FP_due_to_low_iou_correct_class'], 0)
        self.assertEqual(result['FP_due_to_low_iou_wrong_class'], 0)
        self.assertEqual(result['FP_due_to_extra_pred_boxes'], 0)


class TestDetectionMicroPrecisionRecallF1(unittest.TestCase):
    # We include a subset of cases here since the underlying logic is inherited from DetectionConfusionMatrix.
    def test_true_positive(self):
        predictions = [[[0, 1.0, 0, 0, 1, 1]], [[1, 1.0, 0, 0, 1, 1]]]
        targets = [[[0, 0, 0, 1, 1]], [[1, 0, 0, 1, 1]]]

        metric = DetectionMicroPrecisionRecallF1(iou_threshold=0.5)
        metric.update(predictions, targets)
        result = metric.compute()

        self.assertEqual(result['Precision'], 1.0)
        self.assertEqual(result['Recall'], 1.0)
        self.assertEqual(result['F1'], 1.0)

    def test_false_positive_wrong_class(self):
        predictions = [[[0, 1.0, 0, 0, 1, 1]], [[1, 1.0, 0, 0, 1, 1]]]
        targets = [[[0, 0, 0, 1, 1]], [[0, 0, 0, 1, 1]]]

        metric = DetectionMicroPrecisionRecallF1(iou_threshold=0.5)
        metric.update(predictions, targets)
        result = metric.compute()

        self.assertEqual(result['Precision'], 0.5)
        self.assertEqual(result['Recall'], 0.5)
        self.assertEqual(result['F1'], 0.5)

    def test_false_negative(self):
        predictions = [[[0, 1.0, 0, 0, 1, 1]]]
        targets = [[[0, 0, 0, 1, 1], [1, 0, 0, 1, 1]]]

        metric = DetectionMicroPrecisionRecallF1(iou_threshold=0.5)
        metric.update(predictions, targets)
        result = metric.compute()

        self.assertEqual(result['Precision'], 1.0)
        self.assertEqual(result['Recall'], 0.5)
        self.assertEqual(result['F1'], 0.6666666666666666)

    def test_empty_predictions(self):
        predictions = [[[]]]
        targets = [[[0, 0, 0, 1, 1]]]

        metric = DetectionMicroPrecisionRecallF1(iou_threshold=0.5)
        metric.update(predictions, targets)
        result = metric.compute()

        self.assertEqual(result['Precision'], 0.0)
        self.assertEqual(result['Recall'], 0.0)
        self.assertEqual(result['F1'], 0.0)

    def test_empty_targets(self):
        predictions = [[[0, 1.0, 0, 0, 1, 1]]]
        targets = [[]]

        metric = DetectionMicroPrecisionRecallF1(iou_threshold=0.5)
        metric.update(predictions, targets)
        result = metric.compute()

        self.assertEqual(result['Precision'], 0.0)
        self.assertEqual(result['Recall'], 0.0)
        self.assertEqual(result['F1'], 0.0)

    def test_two_images_extra_preds(self):
        predictions = [[[0, 1.0, 0, 0, 1, 1], [1, 1.0, 0, 0, 1, 1]], [[1, 1.0, 0, 0, 1, 1]]]
        targets = [[[0, 0, 0, 1, 1]], [[1, 0, 0, 1, 1]]]

        metric = DetectionMicroPrecisionRecallF1(iou_threshold=0.5)
        metric.update(predictions, targets)
        result = metric.compute()

        self.assertEqual(result['Precision'], 0.6666666666666666)
        self.assertEqual(result['Recall'], 1.0)
        self.assertEqual(result['F1'], 0.8)


if __name__ == '__main__':
    unittest.main()



from visionmetrics.caption.azure_openai_model_eval_base import AzureOpenAITextModelCategoricalEvaluatorBase


# Default prompts
DEFAULT_SYSTEM_MESSAGE = """You are an expert evaluator of AI models. Follow the steps below:
1. Carefully compare the prediction generated by the model against the correct ground truth value and determine how much they differ.
Equally consider important semantic components of the response and details and numbers.
Note that in the ground truth values, <|OR|> indicates that only any of the values between the <|OR|> separators needs to be present.
<|AND|> indicates that all of the values between the <|AND|> separators should be present.
2. Assign a correctness score for the predicted value. Only provide a numeric score between 0.0 and 1.0. Do not provide rationale or other text; ONLY write the score.

Example:
Ground truth: The multi-axes chart titled "Healthcare and Health Analysis: Cost, Coverage, and Premiums" presents data across several healthcare categories.
Private Health Insurance leads in cost at $98,800 and covers 9,502 patients with an average premium of $3,760 and 6 insurers.
Outpatient Care costs $70,000, covers 8,502 patients, with $4,500 average premiums and 4 insurers.
Inpatient Care shows $80,000 in costs, 6,200 patients, a $3,500 average premium, and 5 insurers.
Mental Health Care has a $60,000 cost, covers 4,200 patients, and has $4,000 average premiums with the highest number of insurers at 7.
Dental Care has a $54,000 cost for 3,000 patients covered, a $3,000 average premium, and 9 insurers.
Vision Care, at $18,000 cost, covers the fewest patients, 1,500, with a $2,500 average premium and 3 insurers.
Lastly, Prescription Drugs cost $20,000, cover 5,000 patients, and have a $2,600 average premium with 4 insurers.
Prediction: The bar chart visualizes the cost in dollars across different healthcare categories, along with the number of patients covered, average premiums, and number of insurers.
The healthcare categories include Private Health Insurance with 9,500 patients, Outpatient Care with 8,500 patients, Inpatient Care with 6,000 patients,
Mental Health Care with 4,000 patients, Dental Care with 3,000 patients, Vision Care with 1,000 patients, and Prescription Drugs with 5,000 patients.
Costs are represented by blue bars, patients covered by a blue line, average premiums by pink markers connected with lines, and insurers indicated on the secondary y-axis.
Score: 0.6

Ground truth: 3 dogs
Prediction: 4 dogs
Score: 0.2

Ground truth: The Hunger Games: Catching Fire
Prediction: The Hunger Games
Score: 0.7

Ground truth:
Prediction: 20 pounds
Score: 0.0"""

DEFAULT_PROMPT_TEMPLATE = """Ground truth: <|target|>
Prediction: <|prediction|>
Score: """


class AzureOpenAITextModelCategoricalScore(AzureOpenAITextModelCategoricalEvaluatorBase):
    """
    Prompt-tuned evaluator that uses single-turn text-only calls to an Azure OpenAI model for evaluation and returns
    standard categorical metrics (precision, recall, F1, accuracy) based on the scores returned.

    Args:
        prompt_template: string template for the prompt instructing how to score the prediction. Must have two placeholders: <|prediction|> and <|target|>.
        positive_threshold: determines the threshold (inclusive) at which the score assigned by the model to the prediction means the prediction is correct. Range: [0.0, 1.0].
        negative_value: indicates the value (e.g., None, 'null', 0, '', []) for which predicted and target values are considered null. Note that the evaluator strictly requires
                        an exact match for either a prediction or target to be considered negative.

        Other parameters follow the standards of irisml.tasks.create_azure_openai_chat_model.OpenAITextChatModel;
        see https://github.com/microsoft/irisml-tasks-azure-openai/blob/main/irisml/tasks/create_azure_openai_chat_model.py.
    """
    def __init__(self, endpoint: str, deployment_name: str, system_message=DEFAULT_SYSTEM_MESSAGE, prompt_template=DEFAULT_PROMPT_TEMPLATE,
                 temperature=0.0, max_tokens=50, requests_interval=0, positive_threshold=0.5, negative_value='', api_version="2024-09-01-preview"):
        super().__init__(endpoint=endpoint,
                         deployment_name=deployment_name,
                         system_message=system_message,
                         prompt_template=prompt_template,
                         temperature=temperature,
                         max_tokens=max_tokens,
                         requests_interval=requests_interval,
                         positive_threshold=positive_threshold,
                         negative_value=negative_value,
                         api_version=api_version)



from irisml.tasks.create_azure_openai_chat_model import OpenAITextChatModel
import unittest

from visionmetrics.key_value_pair.key_value_pair_eval import KeyValuePairExtractionScore, SupportedKeyWiseMetric


class TestKeyValuePairExtractionEvaluator(unittest.TestCase):
    ENDPOINT = "https://endpoint-name.openai.azure.com/"
    DEPLOYMENT = "gpt-4o"
    API_VERSION = "2024-08-01-preview"  # Set it to be different from the default
    simple_schema = {
        "image_description": {
            "type": "string",
            "description": "Description of the image in a few sentences, with attention to detail."
        },
        "number_of_chinchillas": {
            "type": "integer",
            "description": "Number of chinchillas visible in the image."
        },
        "estimated_temperature": {
            "type": "number",
            "description": "Estimated temperature of the environment in the image, in degrees Celsius."
        },
        "escaped": {
            "type": "boolean",
            "description": "Whether any chinchillas appear to have escaped the cage."
        },
        "activity": {
            "type": "string",
            "description": "The most salient activity of the chinchillas in the image.",
            "enum": ["sleeping", "running", "playing", "fighting", "eating", "drinking", "none"],
            "includeGrounding": True
        },
        "cage_number": {
            "type": "integer",
            "description": "The cage number being shown in the scene.",
            "enum": [1, 2, 3, 4, 5]
        }
    }
    simple_key_metric_map = {
        "image_description": {
            "metric_name": SupportedKeyWiseMetric.Caption_AzureOpenAITextModelCategoricalScore,
            "metric_args": {"endpoint": ENDPOINT, "deployment_name": DEPLOYMENT, "api_version": API_VERSION},
            "key_trace": ["image_description"]
        },
        "number_of_chinchillas": {
            "metric_name": SupportedKeyWiseMetric.Regression_MeanAbsoluteErrorF1Score,
            "metric_args": {"error_threshold": 0.0},
            "key_trace": ["number_of_chinchillas"]
        },
        "estimated_temperature": {
            "metric_name": SupportedKeyWiseMetric.Regression_MeanAbsoluteErrorF1Score,
            "metric_args": {"error_threshold": 0.0},
            "key_trace": ["estimated_temperature"]
        },
        "escaped": {
            "metric_name": SupportedKeyWiseMetric.Classification_MulticlassF1,
            "metric_args": {"num_classes": 3, "average": "micro"},
            "class_map": {"True": 0, "False": 1, "<|other|>": 2},
            "key_trace": ["escaped"]
        },
        "activity": {
            "metric_name": SupportedKeyWiseMetric.Detection_MicroPrecisionRecallF1,
            "metric_args": {"box_format": "xyxy", "coords": "absolute", "iou_threshold": 0.5},
            "class_map": {"sleeping": 0, "running": 1, "playing": 2, "fighting": 3, "eating": 4, "drinking": 5, "none": 6, "<|other|>": 7},
            "key_trace": ["activity"]
        },
        "cage_number": {
            "metric_name": SupportedKeyWiseMetric.Classification_MulticlassF1,
            "metric_args": {"num_classes": 6, "average": "micro"},
            "class_map": {"1": 0, "2": 1, "3": 2, "4": 3, "5": 4, "<|other|>": 5},
            "key_trace": ["cage_number"]
        }
    }

    simple_list_schema = {
        "defects": {
            "type": "array",
            "description": "The defect types present in the image.",
            "items": {
                "type": "string",
                "description": "The type of defect detected",
                "enum": ["scratch", "dent", "discoloration", "crack"]
            }
        }
    }
    simple_list_key_metric_map = {
        "defects": {
            "metric_name": SupportedKeyWiseMetric.Classification_MultilabelF1WithDuplicates,
            "metric_args": {},
            "class_map": {"scratch": 0, "dent": 1, "discoloration": 2, "crack": 3, "<|other|>": 4},
            "key_trace": ["defects"]
        }
    }

    simple_list_grounding_schema = {
        "defects": {
            "type": "array",
            "description": "The defect types present in the image.",
            "items": {
                "type": "string",
                "description": "The type of defect detected",
                "enum": ["scratch", "dent", "discoloration", "crack"],
                "includeGrounding": True
            }
        }
    }
    simple_list_grounding_key_metric_map = {
        "defects": {
            "metric_name": SupportedKeyWiseMetric.Detection_MicroPrecisionRecallF1,
            "metric_args": {"box_format": "xyxy", "coords": "absolute", "iou_threshold": 0.5},
            "class_map": {"scratch": 0, "dent": 1, "discoloration": 2, "crack": 3, "<|other|>": 4},
            "key_trace": ["defects"]
        }
    }

    complex_list_schema = {
        "defects": {
            "type": "array",
            "description": "The defect types present in the image.",
            "items": {
                "type": "object",
                "properties": {
                    "defect_type": {
                        "type": "string",
                        "description": "The type of defect detected",
                        "enum": ["scratch", "dent", "discoloration", "crack"],
                        "includeGrounding": True
                    },
                    "explanation": {
                        "type": "string",
                        "description": "Rationale for the defects identified."
                    }
                }
            }
        }
    }
    complex_list_key_metric_map = {
        "defects": {
            "metric_name": SupportedKeyWiseMetric.Caption_AzureOpenAITextModelCategoricalScore,
            "metric_args": {"endpoint": ENDPOINT, "deployment_name": DEPLOYMENT, "api_version": API_VERSION},
            "key_trace": ["defects"]
        }
    }

    simple_object_schema = {
        "chart_type": {
            "type": "string",
            "description": "The type of chart shown in the image.",
            "enum": ["line", "bar", "pie", "scatter", "waterfall", "histogram", "gantt", "heatmap"]
        },
        "chart_title": {
            "type": "string",
            "description": "The title of the chart, exactly as it appears in the image."
        },
        "chart_axes": {
            "type": "object",
            "description": "Information about the axes of the chart in the image.",
            "properties": {
                "x_axis_title": {
                    "type": "string",
                    "description": "The title of the x-axis."
                },
                "y_axis_title": {
                    "type": "string",
                    "description": "The title of the y-axis."
                },
                "x_axis_units": {
                    "type": "string",
                    "description": "The units of the x-axis."
                },
                "y_axis_units": {
                    "type": "string",
                    "description": "The units of the y-axis."
                }
            }
        }
    }
    simple_object_key_metric_map = {
        "chart_type": {
            "metric_name": SupportedKeyWiseMetric.Classification_MulticlassF1,
            "metric_args": {"num_classes": 9, "average": "micro"},
            "class_map": {"line": 0, "bar": 1, "pie": 2, "scatter": 3, "waterfall": 4, "histogram": 5, "gantt": 6, "heatmap": 7, "<|other|>": 8},
            "key_trace": ["chart_type"]
        },
        "chart_title": {
            "metric_name": SupportedKeyWiseMetric.Caption_AzureOpenAITextModelCategoricalScore,
            "metric_args": {"endpoint": ENDPOINT, "deployment_name": DEPLOYMENT, "api_version": API_VERSION},
            "key_trace": ["chart_title"]
        },
        "chart_axes_x_axis_title": {
            "metric_name": SupportedKeyWiseMetric.Caption_AzureOpenAITextModelCategoricalScore,
            "metric_args": {"endpoint": ENDPOINT, "deployment_name": DEPLOYMENT, "api_version": API_VERSION},
            "key_trace": ["chart_axes", "value", "x_axis_title"]
        },
        "chart_axes_y_axis_title": {
            "metric_name": SupportedKeyWiseMetric.Caption_AzureOpenAITextModelCategoricalScore,
            "metric_args": {"endpoint": ENDPOINT, "deployment_name": DEPLOYMENT, "api_version": API_VERSION},
            "key_trace": ["chart_axes", "value", "y_axis_title"]
        },
        "chart_axes_x_axis_units": {
            "metric_name": SupportedKeyWiseMetric.Caption_AzureOpenAITextModelCategoricalScore,
            "metric_args": {"endpoint": ENDPOINT, "deployment_name": DEPLOYMENT, "api_version": API_VERSION},
            "key_trace": ["chart_axes", "value", "x_axis_units"]
        }
    }

    complex_object_schema = {
        "brand_sentiment": {
            "type": "object",
            "description": "Attributes of sentiment toward brands depicted in the image.",
            "properties": {
                "has_non_contoso_brands": {
                    "type": "boolean",
                    "description": "Whether the image depicts or contains anything about non-Contoso brands."
                },
                "contoso_specific": {
                    "type": "object",
                    "description": "Sentiment related specifically to the company Contoso.",
                    "properties": {
                        "sentiment": {
                            "type": "string",
                            "description": "Sentiment toward the brand as depicted in the image.",
                            "enum": ["very positive", "somewhat positive", "neutral", "somewhat negative", "very negative"]
                        },
                        "logos": {
                            "type": "array",
                            "description": "The types of Contoso logos present in the image.",
                            "items": {
                                "type": "string",
                                "description": "The type of Contoso logo in the image.",
                                "enum": ["text", "grayscale", "rgb"],
                                "includeGrounding": True
                            }
                        }
                    }
                }
            }
        }
    }
    complex_object_key_metric_map = {
        "brand_sentiment_has_non_contoso_brands": {
            "metric_name": SupportedKeyWiseMetric.Classification_MulticlassF1,
            "metric_args": {"num_classes": 3, "average": "micro"},
            "class_map": {"True": 0, "False": 1, "<|other|>": 2},
            "key_trace": ["brand_sentiment", "value", "has_non_contoso_brands"]
        },
        "brand_sentiment_contoso_specific_sentiment": {
            "metric_name": SupportedKeyWiseMetric.Classification_MulticlassF1,
            "metric_args": {"num_classes": 6, "average": "micro"},
            "class_map": {"very positive": 0, "somewhat positive": 1, "neutral": 2, "somewhat negative": 3, "very negative": 4, "<|other|>": 5},
            "key_trace": ["brand_sentiment", "value", "contoso_specific", "value", "sentiment"]
        },
        "brand_sentiment_contoso_specific_logos": {
            "metric_name": SupportedKeyWiseMetric.Detection_MicroPrecisionRecallF1,
            "metric_args": {"box_format": "xyxy", "coords": "absolute", "iou_threshold": 0.5},
            "class_map": {"text": 0, "grayscale": 1, "rgb": 2, "<|other|>": 3},
            "key_trace": ["brand_sentiment", "value", "contoso_specific", "value", "logos"]
        }
    }

    # In each unit test, we don"t unit test the preprocessor name explicitly because we will test it implicitly via the "update" function call tests.
    def test_two_images_simple_schema(self):
        evaluator = KeyValuePairExtractionScore(key_value_pair_schema=self.simple_schema,
                                                endpoint=self.ENDPOINT,
                                                deployment_name=self.DEPLOYMENT,
                                                api_version=self.API_VERSION)
        with unittest.mock.patch.object(OpenAITextChatModel, "forward", return_value=["1.0"]):
            for key in self.simple_key_metric_map:
                for field in self.simple_key_metric_map[key]:
                    self.assertEqual(evaluator.key_metric_map[key][field], self.simple_key_metric_map[key][field])
            evaluator.update(predictions=[{
                                "image_description": {"value": "Two chinchillas eating."},
                                "number_of_chinchillas": {"value": 2},
                                "estimated_temperature": {"value": 21.0},
                                "escaped": {"value": False},
                                "activity": {"value": "eating", "groundings": [[0, 0, 200, 300]]},
                                "cage_number": {"value": 3}
                            }, {
                                "image_description": {"value": "There is possibly one chinchilla in the cage."},
                                "number_of_chinchillas": {"value": 0},
                                "estimated_temperature": {"value": 20.0},
                                "escaped": {"value": False},
                                "activity": {"value": "none", "groundings": [[0, 100, 200, 200]]},
                                "cage_number": {"value": 2}
                            }],
                            targets=[{
                                "image_description": {"value": "Two chinchillas are eating in a cage."},
                                "number_of_chinchillas": {"value": 2},
                                "estimated_temperature": {"value": 21.0},
                                "escaped": {"value": False},
                                "activity": {"value": "eating", "groundings": [[0, 0, 200, 300]]},
                                "cage_number": {"value": 2}
                            }, {
                                "image_description": {"value": "There is one sleeping chinchilla in the far corner of the cage."},
                                "number_of_chinchillas": {"value": 1},
                                "estimated_temperature": {"value": 20.5},
                                "escaped": {"value": False},
                                "activity": {"value": "snoozing", "groundings": [[0, 100, 250, 250]]},
                                "cage_number": {"value": 1}
                            }])
        report = evaluator.compute()
        self.assertEqual(report["KeyWiseScores"]["image_description"]["F1"], 1.0)
        self.assertEqual(report["KeyWiseScores"]["number_of_chinchillas"]["F1"], 0.5)
        self.assertEqual(report["KeyWiseScores"]["estimated_temperature"]["F1"], 0.5)
        self.assertEqual(report["KeyWiseScores"]["escaped"].item(), 1.)
        self.assertEqual(report["KeyWiseScores"]["activity"]["F1"], 0.5)
        self.assertEqual(report["KeyWiseScores"]["cage_number"].item(), 0.)

        self.assertAlmostEqual(report["MicroF1"], 0.5833333333333333)
        self.assertAlmostEqual(report["MacroF1"], 0.5833333333333333)

    def test_simple_list_schema(self):
        evaluator = KeyValuePairExtractionScore(key_value_pair_schema=self.simple_list_schema,
                                                endpoint=self.ENDPOINT,
                                                deployment_name=self.DEPLOYMENT,
                                                api_version=self.API_VERSION)
        with unittest.mock.patch.object(OpenAITextChatModel, "forward", return_value=["1.0"]):
            for key in self.simple_list_key_metric_map:
                for field in self.simple_list_key_metric_map[key]:
                    self.assertEqual(evaluator.key_metric_map[key][field], self.simple_list_key_metric_map[key][field])
            evaluator.update(predictions=[{
                                "defects": {"value": [
                                    {"value": "scratch"},
                                    {"value": "crack"},
                                    {"value": "scratch"}
                                ]}
                            }],
                            targets=[{
                                "defects": {"value": [
                                    {"value": "scratch"},
                                    {"value": "dent"}
                                ]}
                            }])
            report = evaluator.compute()
            self.assertAlmostEqual(report["KeyWiseScores"]["defects"]["Precision"], 0.3333333333333333)
            self.assertEqual(report["KeyWiseScores"]["defects"]["Recall"], 0.5)
            self.assertEqual(report["KeyWiseScores"]["defects"]["F1"], 0.4)

            self.assertEqual(report["MicroF1"], 0.4)
            self.assertAlmostEqual(report["MacroF1"], 0.4)

    def test_simple_list_grounding_schema(self):
        evaluator = KeyValuePairExtractionScore(key_value_pair_schema=self.simple_list_grounding_schema,
                                                endpoint=self.ENDPOINT,
                                                deployment_name=self.DEPLOYMENT,
                                                api_version=self.API_VERSION)
        with unittest.mock.patch.object(OpenAITextChatModel, "forward", return_value=["1.0"]):
            for key in self.simple_list_grounding_key_metric_map:
                for field in self.simple_list_grounding_key_metric_map[key]:
                    self.assertEqual(evaluator.key_metric_map[key][field], self.simple_list_grounding_key_metric_map[key][field])
            evaluator.update(predictions=[{
                                "defects": {"value": [
                                    {"value": "scratch", "groundings": [[0, 0, 10, 10]]},
                                    {"value": "crack", "groundings": [[10, 10, 20, 20]]}
                                ]}
                            }],
                            targets=[{
                                "defects": {"value": [
                                    {"value": "scratch", "groundings": [[0, 0, 10, 15]]},
                                    {"value": "dent", "groundings": [[10, 10, 20, 20]]}
                                ]}
                            }])
            report = evaluator.compute()
            self.assertEqual(report["KeyWiseScores"]["defects"]["F1"], 0.5)

            self.assertEqual(report["MicroF1"], 0.5)
            self.assertAlmostEqual(report["MacroF1"], 0.5)

    def test_complex_list_schema(self):
        evaluator = KeyValuePairExtractionScore(key_value_pair_schema=self.complex_list_schema,
                                                endpoint=self.ENDPOINT,
                                                deployment_name=self.DEPLOYMENT,
                                                api_version=self.API_VERSION)
        with unittest.mock.patch.object(OpenAITextChatModel, "forward", return_value=["0.0"]):
            for key in self.complex_list_key_metric_map:
                for field in self.complex_list_key_metric_map[key]:
                    self.assertEqual(evaluator.key_metric_map[key][field], self.complex_list_key_metric_map[key][field])
            evaluator.update(predictions=[{
                                "defects": {"value": [{
                                    "defect_type": {"value": "scratch", "groundings": [[0, 0, 10, 10]]},
                                    "explanation": {"value": "There is a line on the otherwise smooth metal."}
                                }]}
                            }],
                            targets=[{
                                "defects": {"value": [{
                                    "defect_type": {"value": "dent", "groundings": [[0, 0, 10, 10]]},
                                    "explanation": {"value": "There is a divot in the upper left."}
                                }]}
                            }])
            report = evaluator.compute()
            self.assertEqual(report["KeyWiseScores"]["defects"]["F1"], 0.0)

            self.assertEqual(report["MicroF1"], 0.0)
            self.assertEqual(report["MacroF1"], 0.0)

    def test_simple_object_schema(self):
        evaluator = KeyValuePairExtractionScore(key_value_pair_schema=self.simple_object_schema,
                                                endpoint=self.ENDPOINT,
                                                deployment_name=self.DEPLOYMENT,
                                                api_version=self.API_VERSION)
        with unittest.mock.patch.object(OpenAITextChatModel, "forward", return_value=["1.0"]):
            for key in self.simple_object_key_metric_map:
                for field in self.simple_object_key_metric_map[key]:
                    self.assertEqual(evaluator.key_metric_map[key][field], self.simple_object_key_metric_map[key][field])
            evaluator.update(predictions=[{
                                "chart_type": {"value": "bar"},
                                "chart_title": {"value": "Number of groundhogs that saw their shadow from 2000-2024"},
                                "chart_axes": {"value": {
                                    "x_axis_title": {"value": "Year"},
                                    "y_axis_title": {"value": "Groundhogs"},
                                    "x_axis_units": {"value": "Year"},
                                    "y_axis_units": {"value": "Count"}
                                }}
                            }],
                            targets=[{
                                "chart_type": {"value": "bar"},
                                "chart_title": {"value": "Count of groundhogs that saw their shadow from 2000-2024"},
                                "chart_axes": {"value": {
                                    "x_axis_title": {"value": "Year"},
                                    "y_axis_title": {"value": "Groundhog"},
                                    "x_axis_units": {"value": "Year"},
                                    "y_axis_units": {"value": "Count"}
                                }}
                            }])
            report = evaluator.compute()
            self.assertEqual(report["KeyWiseScores"]["chart_type"].item(), 1.0)
            self.assertEqual(report["KeyWiseScores"]["chart_title"]["F1"], 1.0)
            self.assertEqual(report["KeyWiseScores"]["chart_axes_x_axis_title"]["F1"], 1.0)
            self.assertEqual(report["KeyWiseScores"]["chart_axes_y_axis_title"]["F1"], 1.0)
            self.assertEqual(report["KeyWiseScores"]["chart_axes_x_axis_units"]["F1"], 1.0)
            self.assertEqual(report["KeyWiseScores"]["chart_axes_y_axis_units"]["F1"], 1.0)

            self.assertEqual(report["MicroF1"], 1.0)
            self.assertEqual(report["MacroF1"], 1.0)

    def test_complex_object_schema(self):
        evaluator = KeyValuePairExtractionScore(key_value_pair_schema=self.complex_object_schema,
                                                endpoint=self.ENDPOINT,
                                                deployment_name=self.DEPLOYMENT,
                                                api_version=self.API_VERSION)
        for key in self.complex_object_key_metric_map:
            for field in self.complex_object_key_metric_map[key]:
                self.assertEqual(evaluator.key_metric_map[key][field], self.complex_object_key_metric_map[key][field])
        evaluator.update(predictions=[{
                            "brand_sentiment": {"value": {
                                "has_non_contoso_brands": {"value": True},
                                "contoso_specific": {"value": {
                                    "sentiment": {"value": "somewhat positive"},
                                    "logos": {"value": [
                                        {"value": "text", "groundings": [[0, 0, 100, 100]]},
                                        {"value": "something_else", "groundings": [[100, 5, 50, 50]]}
                                    ]}
                                }}
                            }}
                        }],
                        targets=[{
                            "brand_sentiment": {"value": {
                                "has_non_contoso_brands": {"value": True},
                                "contoso_specific": {"value": {
                                    "sentiment": {"value": "very positive"},
                                    "logos": {"value": [
                                        {"value": "text", "groundings": [[0, 0, 90, 90]]},
                                        {"value": "rgb", "groundings": [[100, 5, 50, 50]]}
                                    ]}
                                }}
                            }}
                        }])
        report = evaluator.compute()
        self.assertEqual(report["KeyWiseScores"]["brand_sentiment_has_non_contoso_brands"].item(), 1.)
        self.assertEqual(report["KeyWiseScores"]["brand_sentiment_contoso_specific_sentiment"].item(), 0.)
        self.assertEqual(report["KeyWiseScores"]["brand_sentiment_contoso_specific_logos"]["F1"], 0.5)

        self.assertAlmostEqual(report["MicroF1"], 0.5)
        self.assertAlmostEqual(report["MacroF1"], 0.5)

    def test_prediction_missing_key(self):
        evaluator = KeyValuePairExtractionScore(key_value_pair_schema=self.complex_object_schema,
                                                endpoint=self.ENDPOINT,
                                                deployment_name=self.DEPLOYMENT,
                                                api_version=self.API_VERSION)
        evaluator.update(predictions=[{
                            "brand_sentiment": {"value": {
                                "contoso_specific": {"value": {
                                    "sentiment": {"value": "very positive"},
                                    "logos": {"value": [
                                        {"value": "text", "groundings": [[0, 0, 100, 100]]}
                                    ]}
                                }}
                            }}
                        }],
                        targets=[{
                            "brand_sentiment": {"value": {
                                "has_non_contoso_brands": {"value": True},
                                "contoso_specific": {"value": {
                                    "sentiment": {"value": "very positive"},
                                    "logos": {"value": [
                                        {"value": "text", "groundings": [[0, 0, 100, 100]]}
                                    ]}
                                }}
                            }}
                        }])
        report = evaluator.compute()
        self.assertEqual(report["KeyWiseScores"]["brand_sentiment_has_non_contoso_brands"].item(), 0.)
        self.assertEqual(report["KeyWiseScores"]["brand_sentiment_contoso_specific_sentiment"].item(), 1.)
        self.assertEqual(report["KeyWiseScores"]["brand_sentiment_contoso_specific_logos"]["F1"], 1.)

        self.assertAlmostEqual(report["MicroF1"], 0.8)
        self.assertAlmostEqual(report["MacroF1"], 0.6666666666666666)

    def test_target_missing_key(self):
        evaluator = KeyValuePairExtractionScore(key_value_pair_schema=self.complex_object_schema,
                                                endpoint=self.ENDPOINT,
                                                deployment_name=self.DEPLOYMENT,
                                                api_version=self.API_VERSION)
        self.assertRaisesRegex(ValueError,
                               r"The key 'brand_sentiment_contoso_specific_logos' does not exist in the target sample "
                               r"'{'brand_sentiment': {'value': {'has_non_contoso_brands': {'value': True}, "
                               r"'contoso_specific': {'value': {'sentiment': {'value': 'very positive'}}}}}}'.",
                               evaluator.update,
                               [{
                                   "brand_sentiment": {"value": {
                                       "has_non_contoso_brands": {"value": True},
                                       "contoso_specific": {"value": {
                                           "sentiment": {"value": "very positive"},
                                           "logos": {"value": [
                                               {"value": "text", "groundings": [[0, 0, 100, 100]]}
                                           ]}
                                        }}
                                    }}
                                }],
                               [{
                                   "brand_sentiment": {"value": {
                                       "has_non_contoso_brands": {"value": True},
                                       "contoso_specific": {"value": {
                                           "sentiment": {"value": "very positive"}
                                        }}
                                    }}
                                }]
                               )

    def test_prediction_invalid_keys(self):
        evaluator = KeyValuePairExtractionScore(key_value_pair_schema=self.complex_object_schema,
                                                endpoint=self.ENDPOINT,
                                                deployment_name=self.DEPLOYMENT,
                                                api_version=self.API_VERSION)
        evaluator.update(predictions=[{
                            "brand_sentiment": {"value": {
                                "has_non_contoso_brands": {"value": True},
                                "contoso_specific": {"value": {
                                    "sentiment": {"value": "very positive"},
                                    "logos": {"value": [
                                        {"value": "text", "groundings": [[0, 0, 100, 100]]}
                                    ]},
                                    "invalid_key": {"value": None}
                                }},
                                "another_invalid_key": {"value": None}
                            }}
                        }],
                        targets=[{
                            "brand_sentiment": {"value": {
                                "has_non_contoso_brands": {"value": True},
                                "contoso_specific": {"value": {
                                    "sentiment": {"value": "very positive"},
                                    "logos": {"value": [
                                        {"value": "text", "groundings": [[0, 0, 100, 100]]}
                                    ]}
                                }}
                            }}
                        }])
        report = evaluator.compute()
        self.assertEqual(report["KeyWiseScores"]["brand_sentiment_has_non_contoso_brands"].item(), 1.)
        self.assertEqual(report["KeyWiseScores"]["brand_sentiment_contoso_specific_sentiment"].item(), 1.)
        self.assertEqual(report["KeyWiseScores"]["brand_sentiment_contoso_specific_logos"]["F1"], 1.)

        self.assertAlmostEqual(report["MicroF1"], 0.75)
        self.assertAlmostEqual(report["MacroF1"], 1.0)

    def test_target_invalid_keys(self):
        evaluator = KeyValuePairExtractionScore(key_value_pair_schema=self.complex_object_schema,
                                                endpoint=self.ENDPOINT,
                                                deployment_name=self.DEPLOYMENT,
                                                api_version=self.API_VERSION)
        self.assertRaisesRegex(ValueError,
                               r"The target sample '{'brand_sentiment': {'value': {'has_non_contoso_brands': {'value': True}, 'contoso_specific': "
                               r"{'value': {'sentiment': {'value': 'very positive'}, 'logos': {'value': \[{'value': 'text', 'groundings': \[\[0, 0, 100, 100\]\]}\]}, "
                               r"'invalid_key': {'value': 2}}}, 'another_invalid_key': {'value': None}}}}' has at least one invalid key "
                               r"not present in the schema: brand_sentiment_contoso_specific_invalid_key, brand_sentiment_another_invalid_key.",
                               evaluator.update,
                               [{
                                   "brand_sentiment": {"value": {
                                       "has_non_contoso_brands": {"value": True},
                                       "contoso_specific": {"value": {
                                           "sentiment": {"value": "very positive"},
                                           "logos": {"value": [
                                               {"value": "text", "groundings": [[0, 0, 100, 100]]}
                                           ]}
                                        }}
                                    }}
                                }],
                               [{
                                   "brand_sentiment": {"value": {
                                       "has_non_contoso_brands": {"value": True},
                                       "contoso_specific": {"value": {
                                           "sentiment": {"value": "very positive"},
                                           "logos": {"value": [
                                               {"value": "text", "groundings": [[0, 0, 100, 100]]}
                                           ]},
                                           "invalid_key": {"value": 2}
                                        }},
                                       "another_invalid_key": {"value": None}
                                    }}
                                }]
                               )

    def test_two_images_nested_object_schema(self):
        evaluator = KeyValuePairExtractionScore(key_value_pair_schema=self.complex_object_schema,
                                                endpoint=self.ENDPOINT,
                                                deployment_name=self.DEPLOYMENT,
                                                api_version=self.API_VERSION)
        for key in self.complex_object_key_metric_map:
            for field in self.complex_object_key_metric_map[key]:
                self.assertEqual(evaluator.key_metric_map[key][field], self.complex_object_key_metric_map[key][field])
        evaluator.update(predictions=[{
                            "brand_sentiment": {"value": {
                                "has_non_contoso_brands": {"value": True},
                                "contoso_specific": {"value": {
                                    "sentiment": {"value": "somewhat positive"},
                                    "logos": {"value": [
                                        {"value": "text", "groundings": [[0, 0, 100, 100]]}
                                    ]}
                                }}
                            }}
                        },
                        {
                            "brand_sentiment": {"value": {
                                "has_non_contoso_brands": {"value": False},
                                "contoso_specific": {"value": {
                                    "sentiment": {"value": "very positive"},
                                    "logos": {"value": [
                                        {"value": "text", "groundings": [[20, 20, 20, 20]]}
                                    ]}
                                }}
                            }}
                        }],
                        targets=[{
                            "brand_sentiment": {"value": {
                                "has_non_contoso_brands": {"value": True},
                                "contoso_specific": {"value": {
                                    "sentiment": {"value": "somewhat positive"},
                                    "logos": {"value": [
                                        {"value": "text", "groundings": [[0, 0, 100, 100]]}
                                    ]}
                                }}
                            }}
                        },
                        {
                            "brand_sentiment": {"value": {
                                "has_non_contoso_brands": {"value": False},
                                "contoso_specific": {"value": {
                                    "sentiment": {"value": "very positive"},
                                    "logos": {"value": [
                                        {"value": "text", "groundings": [[20, 20, 20, 20]]}
                                    ]}
                                }}
                            }}
                        }])
        report = evaluator.compute()
        self.assertEqual(report["KeyWiseScores"]["brand_sentiment_has_non_contoso_brands"].item(), 1.)
        self.assertEqual(report["KeyWiseScores"]["brand_sentiment_contoso_specific_sentiment"].item(), 1.)
        self.assertEqual(report["KeyWiseScores"]["brand_sentiment_contoso_specific_logos"]["F1"], 1.)

        self.assertAlmostEqual(report["MicroF1"], 1.0)
        self.assertAlmostEqual(report["MacroF1"], 1.0)

    def test_two_images_prediction_invalid_keys(self):
        evaluator = KeyValuePairExtractionScore(key_value_pair_schema=self.complex_object_schema,
                                                endpoint=self.ENDPOINT,
                                                deployment_name=self.DEPLOYMENT,
                                                api_version=self.API_VERSION)
        for key in self.complex_object_key_metric_map:
            for field in self.complex_object_key_metric_map[key]:
                self.assertEqual(evaluator.key_metric_map[key][field], self.complex_object_key_metric_map[key][field])
        evaluator.update(predictions=[{
                            "brand_sentiment": {"value": {
                                "has_non_contoso_brands": {"value": True},
                                "contoso_specific": {"value": {
                                    "sentiment": {"value": "somewhat positive"},
                                    "invalid_key_1": {"value": None},
                                    "logos": {"value": [
                                        {"value": "text", "groundings": [[0, 0, 100, 100]]}
                                    ]}
                                }},
                                "invalid_key_2": {"value": None}
                            }}
                        }, {
                            "brand_sentiment": {"value": {
                                "invalid_key_3": {"value": None},
                                "has_non_contoso_brands": {"value": False},
                                "contoso_specific": {"value": {
                                    "sentiment": {"value": "very positive"},
                                    "logos": {"value": [
                                        {"value": "text", "groundings": [[20, 20, 20, 20]]}
                                    ]},
                                    "invalid_key_4": {"value": None}
                                }}
                            }},
                            "invalid_key_5": {"value": None}
                        }],
                        targets=[{
                            "brand_sentiment": {"value": {
                                "has_non_contoso_brands": {"value": True},
                                "contoso_specific": {"value": {
                                    "sentiment": {"value": "somewhat positive"},
                                    "logos": {"value": [
                                        {"value": "text", "groundings": [[0, 0, 100, 100]]}
                                    ]}
                                }}
                            }}
                        }, {
                            "brand_sentiment": {"value": {
                                "has_non_contoso_brands": {"value": False},
                                "contoso_specific": {"value": {
                                    "sentiment": {"value": "very positive"},
                                    "logos": {"value": [
                                        {"value": "text", "groundings": [[20, 20, 20, 20]]}
                                    ]}
                                }}
                            }}
                        }])
        report = evaluator.compute()
        self.assertEqual(report["KeyWiseScores"]["brand_sentiment_has_non_contoso_brands"].item(), 1.)
        self.assertEqual(report["KeyWiseScores"]["brand_sentiment_contoso_specific_sentiment"].item(), 1.)
        self.assertEqual(report["KeyWiseScores"]["brand_sentiment_contoso_specific_logos"]["F1"], 1.)

        self.assertAlmostEqual(report["MicroF1"], 0.7058823529411765)
        self.assertAlmostEqual(report["MacroF1"], 1.0)

    def test_two_images_prediction_wrong_invalid_keys(self):
        evaluator = KeyValuePairExtractionScore(key_value_pair_schema=self.complex_object_schema,
                                                endpoint=self.ENDPOINT,
                                                deployment_name=self.DEPLOYMENT,
                                                api_version=self.API_VERSION)
        for key in self.complex_object_key_metric_map:
            for field in self.complex_object_key_metric_map[key]:
                self.assertEqual(evaluator.key_metric_map[key][field], self.complex_object_key_metric_map[key][field])
        evaluator.update(predictions=[{
                            "brand_sentiment": {"value": {
                                "has_non_contoso_brands": {"value": True},
                                "contoso_specific": {"value": {
                                    "sentiment": {"value": "somewhat positive"},
                                    "invalid_key_1": {"value": None},
                                    "logos": {"value": [
                                        {"value": "text", "groundings": [[0, 0, 100, 100]]}
                                    ]}
                                }},
                                "invalid_key_2": {"value": None}
                            }}
                        }, {
                            "brand_sentiment": {"value": {
                                "has_non_contoso_brands": {"value": True},
                                "contoso_specific": {"value": {
                                    "sentiment": {"value": "somewhat positive"},
                                    "logos": {"value": [
                                        {"value": "text", "groundings": [[20, 20, 40, 40]]}
                                    ]}
                                }}
                            }}
                        }],
                        targets=[{
                            "brand_sentiment": {"value": {
                                "has_non_contoso_brands": {"value": True},
                                "contoso_specific": {"value": {
                                    "sentiment": {"value": "very positive"},
                                    "logos": {"value": [
                                        {"value": "text", "groundings": [[0, 0, 100, 100]]}
                                    ]}
                                }}
                            }}
                        }, {
                            "brand_sentiment": {"value": {
                                "has_non_contoso_brands": {"value": False},
                                "contoso_specific": {"value": {
                                    "sentiment": {"value": "very positive"},
                                    "logos": {"value": [
                                        {"value": "text", "groundings": [[21, 22, 45, 40]]}
                                    ]}
                                }}
                            }}
                        }])
        report = evaluator.compute()
        self.assertEqual(report["KeyWiseScores"]["brand_sentiment_has_non_contoso_brands"].item(), 0.5)
        self.assertEqual(report["KeyWiseScores"]["brand_sentiment_contoso_specific_sentiment"].item(), 0.0)
        self.assertEqual(report["KeyWiseScores"]["brand_sentiment_contoso_specific_logos"]["F1"], 1.0)

        self.assertAlmostEqual(report["MicroF1"], 0.42857142857142855)
        self.assertAlmostEqual(report["MacroF1"], 0.5)

    def test_simple_schema_batch_update(self):
        evaluator = KeyValuePairExtractionScore(key_value_pair_schema=self.simple_schema,
                                                endpoint=self.ENDPOINT,
                                                deployment_name=self.DEPLOYMENT,
                                                api_version=self.API_VERSION)
        with unittest.mock.patch.object(OpenAITextChatModel, "forward", return_value=["1.0"]):
            for key in self.simple_key_metric_map:
                for field in self.simple_key_metric_map[key]:
                    self.assertEqual(evaluator.key_metric_map[key][field], self.simple_key_metric_map[key][field])
            evaluator.update(predictions=[{
                                "image_description": {"value": "Two chinchillas eating."},
                                "number_of_chinchillas": {"value": 2},
                                "estimated_temperature": {"value": 21},
                                "escaped": {"value": False},
                                "activity": {"value": "eating", "groundings": [[0, 0, 200, 300]]},
                                "cage_number": {"value": 3}
                            }],
                            targets=[{
                                "image_description": {"value": "Two chinchillas are eating in a cage."},
                                "number_of_chinchillas": {"value": 2},
                                "estimated_temperature": {"value": 24},
                                "escaped": {"value": False},
                                "activity": {"value": "eating", "groundings": [[0, 0, 210, 250]]},
                                "cage_number": {"value": 2}
                            }])
            report = evaluator.compute()
            self.assertEqual(report["KeyWiseScores"]["image_description"]["F1"], 1.0)
            self.assertEqual(report["KeyWiseScores"]["number_of_chinchillas"]["F1"], 1.0)
            self.assertEqual(report["KeyWiseScores"]["estimated_temperature"]["F1"], 0.0)
            self.assertEqual(report["KeyWiseScores"]["escaped"].item(), 1.0)
            self.assertEqual(report["KeyWiseScores"]["activity"]["F1"], 1.0)
            self.assertEqual(report["KeyWiseScores"]["cage_number"].item(), 0.0)

            self.assertAlmostEqual(report["MicroF1"], 0.6666666666666666)
            self.assertAlmostEqual(report["MacroF1"], 0.6666666666666666)

            evaluator.update(predictions=[{
                                "image_description": {"value": "There is possibly one chinchilla in the cage."},
                                "number_of_chinchillas": {"value": 0},
                                "estimated_temperature": {"value": 20},
                                "escaped": {"value": False},
                                "activity": {"value": "none", "groundings": [[]]},
                                "cage_number": {"value": 2}
                            }],
                            targets=[{
                                "image_description": {"value": "There is one sleeping chinchilla in the far corner of the cage."},
                                "number_of_chinchillas": {"value": 1},
                                "estimated_temperature": {"value": 20},
                                "escaped": {"value": False},
                                "activity": {"value": "sleeping", "groundings": [[0, 0, 400, 400]]},
                                "cage_number": {"value": 1}
                            }])
        report = evaluator.compute()
        self.assertEqual(report["KeyWiseScores"]["image_description"]["F1"], 1.0)
        self.assertEqual(report["KeyWiseScores"]["number_of_chinchillas"]["F1"], 0.5)
        self.assertEqual(report["KeyWiseScores"]["estimated_temperature"]["F1"], 0.5)
        self.assertEqual(report["KeyWiseScores"]["escaped"].item(), 1.)
        self.assertEqual(report["KeyWiseScores"]["activity"]["F1"], 0.5)
        self.assertEqual(report["KeyWiseScores"]["cage_number"].item(), 0.)

        self.assertAlmostEqual(report["MicroF1"], 0.5833333333333333)
        self.assertAlmostEqual(report["MacroF1"], 0.5833333333333333)


if __name__ == '__main__':
    unittest.main()



from enum import Enum
from functools import reduce
import logging
import operator
import torch
from torchmetrics import Metric

# Import relevant visionmetrics modules; even though they appear to be unused to flake8, they are needed for dynamic metric instantiation at runtime.
from visionmetrics import caption, classification, detection, grounding, regression  # noqa: F401
from visionmetrics.common.utils import precision_recall_f1_scalar

logger = logging.getLogger(__name__)


class SupportedKeyWiseMetric(str, Enum):
    Caption_AzureOpenAITextModelCategoricalScore = "caption.AzureOpenAITextModelCategoricalScore"
    Classification_MulticlassAccuracy = "classification.MulticlassAccuracy"
    Classification_MultilabelAccuracy = "classification.MultilabelAccuracy"
    Classification_MulticlassF1 = "classification.MulticlassF1Score"
    Classification_MultilabelF1 = "classification.MultilabelF1Score"
    Classification_MultilabelF1WithDuplicates = "classification.MultilabelF1ScoreWithDuplicates"
    Detection_MeanAveragePrecision = "detection.MeanAveragePrecision"
    Detection_MicroPrecisionRecallF1 = "detection.DetectionMicroPrecisionRecallF1"
    Regression_MeanAbsoluteError = "regression.MeanAbsoluteError"
    Regression_MeanAbsoluteErrorF1Score = "regression.MeanAbsoluteErrorF1Score"


SUPPORTED_KEY_WISE_METRICS = [m.value for m in SupportedKeyWiseMetric]
SUPPORTED_F1_METRICS = [
    SupportedKeyWiseMetric.Caption_AzureOpenAITextModelCategoricalScore,
    SupportedKeyWiseMetric.Classification_MulticlassF1,
    SupportedKeyWiseMetric.Classification_MultilabelF1,
    SupportedKeyWiseMetric.Classification_MultilabelF1WithDuplicates,
    SupportedKeyWiseMetric.Detection_MicroPrecisionRecallF1,
    SupportedKeyWiseMetric.Regression_MeanAbsoluteErrorF1Score
]
METRICS_WITH_TENSOR_INPUT = [SupportedKeyWiseMetric.Classification_MulticlassAccuracy, SupportedKeyWiseMetric.Classification_MulticlassF1,
                             SupportedKeyWiseMetric.Classification_MultilabelAccuracy, SupportedKeyWiseMetric.Classification_MultilabelF1,
                             SupportedKeyWiseMetric.Regression_MeanAbsoluteError, SupportedKeyWiseMetric.Regression_MeanAbsoluteErrorF1Score]


class KeyValuePairEvaluatorBase(Metric):
    """
    Key-value pair extraction refers to the task of arbitrary schema-based structured field extraction. Each schema is an adapted version of a
    JSON Schema (https://json-schema.org/understanding-json-schema/reference)-formatted dictionary that contains keys, each of which specifies
    the standard JSON Schema type of the key and a string description, whether to perform grounding on the key, classes for closed-vocabulary
    values, and additional information describing list items and object properties (sub-keys).

    Based on a key_metric_map dictionary specifying the mapping between each key of such a schema and the corresponding visionmetrics metric, metric arguments,
    preprocessors for predictions and targets, and a list trace describing the path to the key through the prediction and target objects,
    this evaluator computes the key-wise metrics for each key in the schema and returns the overall micro F1, macro F1, and key-wise scores in their raw format for each key.
    Each key can have a different metric. The metrics supported are specified in SupportedKeyWiseMetric.

    For each key, definitions of true positive, false positive, and false negative are inherited from the corresponding metric.
    In addition to metric-specific definitions, missing keys in predictions are counted as false negatives, and invalid keys in predictions are counted as false positives.

    Args:
        key_metric_map: dictionary from keys (extracted field names) to a dictionary with four required fields:
        1. metric_name: string of the metric name as defined in visionmetrics (e.g., 'classification.MulticlassAccuracy'), which should be among the ones specified in SupportedKeyWiseMetric.
        2. metric_args: dictionary of args to pass in as keyword arguments to the initialization function of the metric.
        3. preprocessor: function object that can be called with (prediction, target) values for a single instance to preprocess them into the desired format for the corresponding metric.
        4. key_trace: list of strings of key names that traces the path to the current key in the key-value pair prediction/target object (not in the schema).
        Examples (corresponding to the examples in key_value_pair_eval.py):
        defect_detection_key_metric_map = {
            "defects": {
                "metric_name": SupportedKeyWiseMetric.Detection_MicroPrecisionRecallF1,
                "metric_args": {"box_format": "xyxy", "coords": "absolute", "iou_threshold": 0.5},
                "prediction_preprocessor": <reference to detection prediction preprocessing function; see example implementation in key_value_pair_eval.py>,
                "target_preprocessor": <reference to detection target preprocessing function; see example implementation in key_value_pair_eval.py>,
                "key_trace": ["defects"]
            },
            "rationale": {
                "metric_name": SupportedKeyWiseMetric.Caption_AzureOpenAITextModelCategoricalScore,
                "metric_args": {"endpoint": <endpoint>, "deployment_name": <deployment>},
                "prediction_preprocessor": <reference to captioning prediction preprocessing function; see example implementations in key_value_pair_eval.py>,
                "target_preprocessor": <reference to captioning target preprocessing function; see example implementations in key_value_pair_eval.py>,
                "key_trace": ["rationale"]
            }
        }
        brand_sentiment_key_metric_map = {
            "brand_sentiment_has_non_contoso_brands": {
                "metric_name": SupportedKeyWiseMetric.Classification_MulticlassF1,
                "metric_args": {"num_classes": 3, "average": "micro"},
                "prediction_preprocessor": <reference to multiclass classification prediction preprocessing function; see example implementation in key_value_pair_eval.py>,
                "target_preprocessor": <reference to multiclass classification target preprocessing function; see example implementation in key_value_pair_eval.py>,
                "key_trace": ["brand_sentiment", "value", "has_non_contoso_brands"]
            },
            "brand_sentiment_contoso_specific_sentiment": {
                "metric_name": SupportedKeyWiseMetric.Classification_MulticlassF1,
                "metric_args": {"num_classes": 6, "average": "micro"},
                "prediction_preprocessor": <reference to multiclass classification prediction preprocessing function; see example implementation in key_value_pair_eval.py>,
                "target_preprocessor": <reference to multiclass classification target preprocessing function; see example implementation in key_value_pair_eval.py>,
                "key_trace": ["brand_sentiment", "value", "contoso_specific", "value", "sentiment"]
            },
            "brand_sentiment_contoso_specific_logos": {
                "metric_name": SupportedKeyWiseMetric.Detection_MicroPrecisionRecallF1,
                "metric_args": {"box_format": "xyxy", "coords": "absolute", "iou_threshold": 0.5},
                "prediction_preprocessor": <reference to detection prediction preprocessing function; see example implementation in key_value_pair_eval.py>,
                "target_preprocessor": <reference to detection target preprocessing function; see example implementation in key_value_pair_eval.py>,
                "key_trace": ["brand_sentiment", "value", "contoso_specific", "value", "logos"]
            }
        }
    }
    """
    def __init__(self, key_metric_map: dict):
        super().__init__()
        if not isinstance(key_metric_map, dict):
            raise ValueError("key_metric_map must be a dictionary mapping keys to their corresponding evaluation metrics.")
        self.key_metric_map = key_metric_map

        self.key_evaluator_map = {}
        for key in key_metric_map:
            # Validate metric_name and metric_args
            try:
                metric_name = key_metric_map[key]["metric_name"]
                metric_args = key_metric_map[key]["metric_args"]
            except KeyError:
                raise ValueError(f"Each value in key_metric_map must be a dictionary with a 'metric_name' key for a supported metric in visionmetrics of the form <task_name>.<metric_name>"
                                 f" and a 'metric_args' key for a dictionary of arguments to pass to the initialization function of that metric. Key '{key}' does not satisfy this rule.")
            if isinstance(metric_name, SupportedKeyWiseMetric):
                metric_name = metric_name.value
            if metric_name not in SUPPORTED_KEY_WISE_METRICS:
                raise ValueError(f"Metric '{key_metric_map[key]['metric_name']}' is not supported. "
                                 f"Each key's metric must be among the supported metrics: {', '.join(SUPPORTED_KEY_WISE_METRICS)}.")

            # Instantiate metric
            try:
                self.key_evaluator_map[key] = eval(f"{metric_name}(**metric_args)")
            except Exception as e:
                raise ValueError(f"Encountered error '{e}' when instantiating metric '{metric_name}' for key '{key}' with arguments '{metric_args}'.")
        self.invalid_predicted_keys = []
        self.missing_predicted_keys = []
        logger.info(f"Initialized KeyValuePairEvaluatorBase with key_metric_map {key_metric_map} and key_evaluator_map {self.key_evaluator_map}.")

    def _get_invalid_keys(self, sample, key_trace: list = [], invalid_keys: list = []):
        if isinstance(sample, dict):
            for key in sample:
                self._get_invalid_keys(sample[key]["value"], key_trace=key_trace + [key], invalid_keys=invalid_keys)
        else:
            flattened_key_name = '_'.join(key_trace)
            if flattened_key_name not in self.key_metric_map:
                invalid_keys.append(flattened_key_name)

    def update(self, predictions, targets):
        """
        Updates metrics for each key using all samples in predictions and targets.
        Both predictions and targets should be dictionaries of the form {'<key>': <value>}, where <value> is in the format expected for the respective metric for that key.
        Each sample in predictions and targets must have the same keys (though they do not have to have all the keys in the dataset).
        """
        # Check for invalid keys in both target (throws an error) and prediction (counted as false positives)
        for prediction, target in zip(predictions, targets):
            target_invalid_keys = []
            self._get_invalid_keys(sample=target, invalid_keys=target_invalid_keys)
            if target_invalid_keys:
                raise ValueError(f"The target sample '{target}' has at least one invalid key not present in the schema: {', '.join(target_invalid_keys)}.")
            self._get_invalid_keys(sample=prediction, invalid_keys=self.invalid_predicted_keys)
        if len(self.invalid_predicted_keys) > 0:
            logger.debug(f"Invalid keys were present in the predictions in this update: {', '.join(self.invalid_predicted_keys)}.")

        for key, metric in self.key_evaluator_map.items():
            metric_name = self.key_metric_map[key]["metric_name"]
            key_trace = self.key_metric_map[key]["key_trace"]
            key_predictions = []
            key_targets = []
            for prediction, target in zip(predictions, targets):
                # Use the key trace to traverse the prediction and target objects to get the values for these keys
                try:
                    key_prediction = reduce(operator.getitem, key_trace, prediction)
                except KeyError:
                    self.missing_predicted_keys.append(key)
                    logger.debug(f"The key '{key}' does not exist in the prediction sample '{prediction}'.")
                    continue
                try:
                    key_target = reduce(operator.getitem, key_trace, target)
                except KeyError:
                    raise ValueError(f"The key '{key}' does not exist in the target sample '{target}'.")

                # Construct expected evaluation metric update format for the current key
                try:
                    prediction_preprocessor = self.key_metric_map[key]["prediction_preprocessor"]
                    key_prediction_formatted = prediction_preprocessor(key_prediction)
                    key_predictions.append(key_prediction_formatted)
                except ValueError as e:
                    logger.debug(f"Encountered error {e} when preprocessing prediction '{key_prediction}' for key '{key}' to the '{self.key_metric_map[key]['metric_name']}' metric's"
                                 " expected format.")
                try:
                    target_preprocessor = self.key_metric_map[key]["target_preprocessor"]
                    key_target_formatted = target_preprocessor(key_target)
                    key_targets.append(key_target_formatted)
                except ValueError as e:
                    logger.debug(f"Encountered error {e} when preprocessing target '{key_target}' for key '{key}' to the '{self.key_metric_map[key]['metric_name']}' metric's expected format.")

            # Convert lists to tensors for metrics that expect torch tensors
            if metric_name in METRICS_WITH_TENSOR_INPUT:
                key_predictions = torch.tensor(key_predictions)
                key_targets = torch.tensor(key_targets)

            try:
                if len(key_predictions) > 0 and len(key_targets) > 0:
                    metric.update(key_predictions, key_targets)
            except Exception as e:
                raise ValueError(f"Encountered error '{e}' when updating metric '{self.key_metric_map[key]['metric_name']}' for key '{key}'.")

    def compute(self):
        """
        Computes key-wise metrics and returns a dictionary mapping keys to verbatim results from the evaluator for the corresponding key.
        """
        key_wise_scores = {k: None for k in self.key_evaluator_map}
        for key in self.key_evaluator_map:
            try:
                metric = self.key_evaluator_map[key]
                key_wise_scores[key] = metric.compute()
            except Exception as e:
                raise ValueError(f"Encountered error '{e}' when computing metric '{self.key_metric_map[key]['metric_name']}' for key '{key}'.")

        # If all keys support F1 scores, compute total micro- and macro-averaged F1 scores
        if all([self.key_metric_map[key]["metric_name"] in SUPPORTED_F1_METRICS]):
            macro_f1 = 0.
            total_tp, total_fp, total_fn = 0, 0, 0
            for key in self.key_metric_map:
                metric_name = self.key_metric_map[key]["metric_name"]
                if metric_name == SupportedKeyWiseMetric.Caption_AzureOpenAITextModelCategoricalScore:
                    macro_f1 += key_wise_scores[key]["F1"]
                    total_tp += key_wise_scores[key]["TP"]
                    total_fp += key_wise_scores[key]["FP_GTNull"]
                    total_fp += key_wise_scores[key]["FP_GTNotNull"]
                    total_fn += key_wise_scores[key]["FN"]
                elif metric_name == SupportedKeyWiseMetric.Classification_MulticlassF1 or metric_name == SupportedKeyWiseMetric.Classification_MultilabelF1:
                    macro_f1 += key_wise_scores[key].item()
                    total_tp += self.key_evaluator_map[key].tp.sum().item()
                    total_fp += self.key_evaluator_map[key].fp.sum().item()
                    total_fn += self.key_evaluator_map[key].fn.sum().item()
                elif metric_name in [SupportedKeyWiseMetric.Detection_MicroPrecisionRecallF1,
                                     SupportedKeyWiseMetric.Regression_MeanAbsoluteErrorF1Score,
                                     SupportedKeyWiseMetric.Classification_MultilabelF1WithDuplicates]:
                    macro_f1 += key_wise_scores[key]["F1"]
                    total_tp += self.key_evaluator_map[key].tp.item()
                    total_fp += self.key_evaluator_map[key].fp.item()
                    total_fn += self.key_evaluator_map[key].fn.item()
            # Note: Currently, macro_f1 does not take into account missing or invalid keys, only those present in both the target and prediction for each sample.
            macro_f1 = macro_f1 / len(self.key_metric_map)
            total_fp += len(self.invalid_predicted_keys)
            total_fn += len(self.missing_predicted_keys)
            precision_recall_f1 = precision_recall_f1_scalar(tp=total_tp, fp=total_fp, fn=total_fn)

        return {
            "MicroF1": precision_recall_f1["F1"],
            "MacroF1": macro_f1,
            "KeyWiseScores": key_wise_scores
        }



# Imports


def precision_recall_f1_scalar(tp: int, fp: int, fn: int):
    """
    Computes precision, recall, and F1 from scalar values, defined as zero when there is division by zero.
    Args:
        tp: int, the number of true positives.
        fp: int, the number of false positives.
        fn: int, the number of false negatives.
    """
    try:
        precision = tp / (tp + fp)
    except ZeroDivisionError:
        precision = 0.
    try:
        recall = tp / (tp + fn)
    except ZeroDivisionError:
        recall = 0.
    try:
        f1 = (2 * precision * recall) / (precision + recall)
    except ZeroDivisionError:
        f1 = 0.
    return {
        "Precision": precision,
        "Recall": recall,
        "F1": f1
    }
